""""code to fit learning curves
adapted from
https://github.com/NickleDave/learning-curves/"""

import numpy as np
from scipy import optimize


def residual_two_functions(params, x, y1, y1err, y2, y2err):
    """
    returns residuals
    between two lines, specified by parameters in variable params,
    and data y1 and y2
    """

    b = params[0]
    alpha = params[1]
    c = params[2]
    beta = params[3]
    asymptote = params[4]
    diff1 = (y1 - (asymptote + b * alpha ** x)) ** 2 / y1err
    diff2 = (y2 - (asymptote + c * beta ** x)) ** 2 / y2err
    return np.concatenate((diff1, diff2))


def fit_learning_curve(train_set_size, error_test, error_train=None,
                       pinit=(1.0, -1.0), funcs=1):
    """
    returns parameters to predict learning curve as a power function with the form
    y = a + b * x**alpha
    where x is the training set size, i.e., the independent variable

    You provide the function with your data: a vector of the training set sizes you used, and arrays of the error
    you found when training models with those training sets. The function then returns the fit parameters.
    Based on [1]_.

    Parameters
    ----------
    train_set_size : ndarray
        vector of m integers representing number of samples
        in training sets, should increase monotonically
    error_test : ndarray
        m x n array of errors where error_train[m,n] is
        the error measured for replicate n of training a model
        with train_set_size[m] samples.
        Error is measured on on a test set separate from the training set.
    error_train : ndarray
        same as error_test except the error is measured on the *training* set.
        Default is None.
    pinint : list
        initial guess for parameters b and alpha, default is [1.0, -1.0]
    funcs : int
        number of functions to fit, default is 1.
        If funcs==1 and only test error is passed as an argument,
        a power function is fit just to the test error
        If funcs==1 and both test error and train error are passed as arguments,
        it is assumed the train error and test error can be fit with same
        exponent and scaling parameter.
        If funcs==2, both test error and train error must be passed
        and each is fit with separate exponent and scaling parameters,
        but both share an extra parameter which is the asymptote.

    Returns
    -------
    a: float
        asymptotic value of error predicted for infinite training data
    b: float
        scaling parameter of power function
    alpha: float
        exponent parameter of power function

    *** if funcs = 2 ***
    c: float
        scaling parameter of power function fit to train error (b fits test error)
    beta: float
        exponent parameter of power function fit to train error (alpha fits test error)

    .. [1] Cortes, Corinna, et al.
    "Learning curves: Asymptotic values and rate of convergence."
    Advances in Neural Information Processing Systems. 1994.
    """

    if funcs not in [1, 2]:
        raise ValueError('funcs argument should equal 1 or 2')

    if funcs == 2 and error_train is None:
        raise ValueError('error_train is a required argument when funcs==2')

    if train_set_size.shape[0] != error_test.shape[0]:
        raise ValueError(
            'Number of elements in train_set_size does not match number of columns in error_test')

    fitfunc = lambda p, x: p[0] + p[1] * x
    errfunc = lambda p, x, y, err: (y - fitfunc(p, x)) / err

    logx = np.log10(train_set_size)

    if error_train is None:  # if we just have test error, fit with power function
        y = np.mean(error_test, axis=1)
        logy = np.log10(y)
        yerr = np.std(error_test, axis=1)
        logyerr = yerr / y
        out1 = optimize.leastsq(errfunc, pinit,
                                args=(logx, logy, logyerr), full_output=True)
        pfinal = out1[0]
        b = 10.0 ** pfinal[0]
        alpha = pfinal[1]
        return b, alpha

    elif error_train is not None and funcs == 1:  # if we have train error too, then try Cortes et al. 1994 approach
        err_diff = error_test - error_train
        y = np.mean(err_diff, axis=1)
        logy = np.log10(y)
        yerr = np.std(err_diff, axis=1)
        logyerr = yerr / y
        out1 = optimize.leastsq(errfunc, pinit,
                                args=(logx, logy, logyerr), full_output=True)
        pfinal = out1[0]
        b = (10.0 ** pfinal[0]) / 2
        alpha = pfinal[1]

        err_sum = error_test + error_train
        y2 = np.mean(err_sum, axis=1)
        logy2 = np.log10(y2)
        y2err = np.std(err_sum, axis=1)
        logy2err = y2err / y
        # take mean of logy as best estimate of horizontal line
        estimate = np.average(logy2, weights=logy2err)
        a = (10.0 ** estimate) / 2
        return a, b, alpha

    elif error_train is not None and funcs == 2:
        y1 = np.mean(error_test, axis=1)
        y1err = np.std(error_test, axis=1)
        logy1 = np.log10(y1)
        y2 = np.mean(error_train, axis=1)
        y2err = np.std(error_train, axis=1)
        logy2 = np.log10(y2)
        if len(pinit) < 3:  # if default pinit from function declaration
            # change instead to default pinit in next line
            pinit = [1.0, -1.0, 1.0, 1.0, 0.05]
        best, cov, info, message, ier = optimize.leastsq(residual_two_functions,
                                                         pinit,
                                                         args=(train_set_size, y1, y1err, y2, y2err),
                                                         full_output=True)
        return best




"""parses [PREDICT] section of config"""
from configparser import NoOptionError

import attr
from attr.validators import instance_of, optional

from .validators import is_a_directory, is_a_file
from .. import network


@attr.s
class PredictConfig:
    """class that represents [PREDICT] section of config.ini file

    Attributes
    ----------
    checkpoint_path : str
        path to directory with checkpoint files saved by Tensorflow, to reload model
    networks : namedtuple
        where each field is the Config tuple for a neural network and the name
        of that field is the name of the class that represents the network.
    labels_mapping_path : str
        path to file that contains labels mapping, to convert output from consecutive
        digits back to labels used for audio segments (e.g. birdsong syllables)
    dir_to_predict : str
        path to directory where input files are located
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    """
    checkpoint_path = attr.ib(validator=[instance_of(str), is_a_directory])
    networks = attr.ib()
    labels_mapping_path = attr.ib(validator=[instance_of(str), is_a_file])
    dir_to_predict = attr.ib(validator=[instance_of(str), is_a_directory])
    spect_scaler_path = attr.ib(validator=optional([instance_of(str), is_a_file]), default=None)


def parse_predict_config(config):
    """parse [PREDICT] section of config.ini file

    Parameters
    ----------
    config : ConfigParser
        containing config.ini file already loaded by parse function

    Returns
    -------
    predict_config : vak.config.predict.PredictConfig
        instance of PredictConfig class that represents [PREDICT] section
        of config.ini file
    """
    config_dict = {}

    try:
        config_dict['checkpoint_path'] = config['PREDICT']['checkpoint_path']
    except NoOptionError:
        raise KeyError('must specify checkpoint_path in [PREDICT] section '
                       'of config.ini file')

    # load entry points within function, not at module level,
    # to avoid circular dependencies
    # (user would be unable to import networks in other packages
    # that subclass vak.network.AbstractVakNetwork
    # since the module in the other package would need to `import vak`)
    NETWORKS = network._load()
    NETWORK_NAMES = NETWORKS.keys()
    try:
        networks = [network_name for network_name in
                    config['PREDICT']['networks'].split(',')]
        for network_name in networks:
            if network_name not in NETWORK_NAMES:
                raise TypeError('Neural network {} not found when importing installed networks.'
                                .format(network))
        config_dict['networks'] = networks
    except NoOptionError:
        raise KeyError("'networks' option not found in [PREDICT] section of config.ini file. "
                       "Please add this option as a comma-separated list of neural network names, e.g.:\n"
                       "networks = TweetyNet, GRUnet, convnet")

    try:
        config_dict['labels_mapping_path'] = config['PREDICT']['labels_mapping_path']
    except NoOptionError:
        raise KeyError('must specify labels_mapping_path in [PREDICT] section '
                       'of config.ini file')

    try:
        config_dict['dir_to_predict'] = config['PREDICT']['dir_to_predict']
    except NoOptionError:
        raise KeyError('must specify dir_to_predict in [PREDICT] section '
                       'of config.ini file')

    if config.has_option('PREDICT','spect_scaler_path'):
        config_dict['spect_scaler_path'] = config['PREDICT']['spect_scaler_path']

    return PredictConfig(**config_dict)



"""parses [SPECTROGRAM] section of config"""
from collections import namedtuple


SpectConfig = namedtuple('SpectConfig', ['fft_size',
                                         'step_size',
                                         'freq_cutoffs',
                                         'thresh',
                                         'transform_type'])


def parse_spect_config(config):
    """parse [SPECTROGRAM] section of config.ini file

    Parameters
    ----------
    config : ConfigParser
        containing config.ini file already loaded by parse function

    Returns
    -------
    spect_config : namedtuple
        with fields:
            fft_size
            step_size
            freq_cutoffs
            thresh
            transform_type
    """
    fft_size = int(config['SPECTROGRAM']['fft_size'])
    step_size = int(config['SPECTROGRAM']['step_size'])

    if config.has_option('SPECTROGRAM', 'freq_cutoffs'):
        freq_cutoffs = [float(element)
                        for element in
                        config['SPECTROGRAM']['freq_cutoffs'].split(',')]
        if len(freq_cutoffs) != 2:
            raise ValueError('freq_cutoffs should be a list of two elements, but instead'
                             'got: {}'.format(freq_cutoffs))
        if freq_cutoffs[0] > freq_cutoffs[1]:
            raise ValueError('lower freq_cutoff should be less than higher freq_cutoff,'
                             'instead of: {}'.format(freq_cutoffs))
    else:
        freq_cutoffs = None

    if config.has_option('SPECTROGRAM', 'thresh'):
        thresh = float(config['SPECTROGRAM']['thresh'])
    else:
        thresh = None

    if config.has_option('SPECTROGRAM', 'transform_type'):
        transform_type = config['SPECTROGRAM']['transform_type']
        valid_transform_types = {'log_spect', 'log_spect_plus_one'}
        if config['SPECTROGRAM']['transform_type'] not in valid_transform_types:
            raise ValueError('Value for `transform_type`, {}, in [SPECTROGRAM] '
                             'section of .ini file is not recognized. Must be one '
                             'of the following: {}'
                             .format(spect_params['transform_type'],
                                     valid_transform_types))
    else:
        transform_type = None

    return SpectConfig(fft_size,
                       step_size,
                       freq_cutoffs,
                       thresh,
                       transform_type)



"""module that contains helper function to load networks, and
the AbstractVakNetwork class.
Networks in separate packages should subclass this class, and
then make themselves available to vak by including
'vak.network' in the entry_points value of their setup.py
file.

The reason for subclassing is simply to set a
standard interface for networks to use. vak will check
that any network implements all the methods listed here,
since it requires those methods to carry out commands
in its command-line interface.

For example, if you had a package `grunet` containing a class
`GRUnet`, then that package would include the following in its
setup.py file:

setup(
    ...
    entry_points={'vak.network': 'GRUnet = grunet.model:GRUnet'},
    ...
)

For more detail on entry points in Python, see:
https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata
https://setuptools.readthedocs.io/en/latest/setuptools.html#dynamic-discovery-of-services-and-plugins
https://amir.rachum.com/blog/2017/07/28/python-entry-points/
"""
import typing
import pkg_resources


def _load():
    networks = {
        entry_point.name: entry_point.load()
        for entry_point
        in pkg_resources.iter_entry_points('vak.network')
    }
    return networks


class AbstractVakNetwork:

    Config = typing.NamedTuple('Config',
                               [('conv_layers', int),
                                ('kernels', int),
                                ('learning_rate', float)])

    def __init__(self):
        pass

    def inference(self):
        """inference method, forward pass through network
        with output that is used with loss function for
        training (e.g., softmax layer)"""
        pass

    def optimize(self):
        """optimize method, that runs optimizer on one cost value, given
        ground truth as expected output"""
        pass

    def error(self):
        """error method, that returns mean error given as input a spectrogram
        and the true and predicted labels"""

    def train(self):
        """method that uses error and optimize methods to iteratively
        train the network"""
        pass

    def predict(self):
        """predict method, that (usually) returns argmax(inference)
        for all outputs"""
        pass

    def save(self):
        """method to save network"""
        pass

    def restore(self):
        pass



"""functions to plot learning curve results"""
from glob import glob
import os
import pickle
from configparser import ConfigParser

import joblib
import numpy as np
import matplotlib.pyplot as plt
%matplotlib inline


def get_all_results_list(root, these_dirs_branch, config_file):
    these_dirs = [(root + this_dir[0], this_dir[1]) for this_dir in these_dirs_branch]
    config = ConfigParser()

    all_results_list = []

    for this_dir, bird_ID in these_dirs:
        results_dict = {}
        results_dict['dir'] = this_dir
        os.chdir(this_dir)
        config.read(config_file)
        results_dict['data_dir'] = config['DATA']['data_dir']
        results_dict['time_steps'] = config['NETWORK']['time_steps']
        results_dict['num_hidden'] = config['NETWORK']['num_hidden']
        results_dict['train_set_durs'] = [int(element)
                                          for element in
                                          config['TRAIN']['train_set_durs'].split(',')]
        with open(glob('summary*/train_err')[0], 'rb') as f:
            results_dict['train_err'] = pickle.load(f)
        with open(glob('summary*/test_err')[0], 'rb') as f:
            results_dict['test_err'] = pickle.load(f)
        pe = joblib.load(glob('summary*/y_preds_and_err_for_train_and_test')[0])
        results_dict['train_syl_err_rate'] = pe['train_syl_err_rate']
        results_dict['test_syl_err_rate'] = pe['test_syl_err_rate']
        results_dict['bird_ID'] = bird_ID
        all_results_list.append(results_dict)
        os.chdir('..')

    for el in all_results_list:
        el['mn_test_err'] = np.mean(el['test_err'], axis=1)
        el['mn_train_err'] = np.mean(el['train_err'], axis=1)
        el['mn_train_syl_err'] = np.mean(el['train_syl_err_rate'], axis=1)
        el['mn_test_syl_err'] = np.mean(el['test_syl_err_rate'], axis=1)
        el['train_set_durs'] = np.asarray(el['train_set_durs'])

    return all_results_list


def frame_error_rate(all_results_list):
    all_mn_test_err = []
    for el in all_results_list:
        all_mn_test_err.append(el['mn_test_err'])
    all_mn_test_err = np.asarray(all_mn_test_err)

    plt.style.use('ggplot')
    fig, ax = plt.subplots()
    for el in all_results_list:
        lbl = (el['bird_ID'])
        ax.plot(el['train_set_durs'],
                el['mn_test_err'],
                label=lbl,
                linestyle='--',
                marker='o')
    ax.plot(el['train_set_durs'], np.median(all_mn_test_err, axis=0),
            linestyle='--', marker='o', linewidth=3, color='k', label='median across birds')

    fig.set_size_inches(16, 8)
    plt.legend(fontsize=20)
    plt.xticks(el['train_set_durs'])
    plt.tick_params(axis='both', which='major', labelsize=20, rotation=45)
    plt.title('Frame error rate as a function of training set size', fontsize=40)
    plt.ylabel('Frame error rate\nas measured on test set', fontsize=32)
    plt.xlabel('Training set size: duration in s', fontsize=32);
    plt.tight_layout()
    plt.savefig('frame-err-rate-v-train-set-size.png')


def syllable_error_rate(all_results_list):
    plt.style.use('ggplot')
    fig, ax = plt.subplots()
    fig.set_size_inches(16, 8)
    for el in all_results_list:
        lbl = (el['bird_ID'])
        ax.plot(el['train_set_durs'],
                el['mn_test_syl_err'],
                label=lbl,
                linestyle=':',
                marker='o')

    plt.scatter(120, 0.84, s=75)
    plt.text(75, 0.7, 'Koumura & Okanoya 2016,\n0.84 note error rate\nwith 120s training data', fontsize=20)
    plt.scatter(480, 0.5, s=75)
    plt.text(355, 0.35, 'Koumura & Okanoya 2016,\n0.46 note error rate\nwith 480s training data', fontsize=20)

    plt.legend(fontsize=20, loc='upper right');
    plt.title('Syllable error rate as a function of training set size', fontsize=40)
    plt.xticks(el['train_set_durs'])
    plt.tick_params(axis='both', which='major', labelsize=20, rotation=45)
    plt.ylabel('Syllable error rate\nas measured on test set', fontsize=32)
    plt.xlabel('Training set size: duration in s', fontsize=32);
    plt.tight_layout()
    plt.savefig('syl-error-rate-v-train-set-size.png')


from configparser import ConfigParser
import logging
import os
import pickle
import shutil
import sys
import warnings
from datetime import datetime

import joblib
import numpy as np
import tensorflow as tf
from tqdm import tqdm

from .. import network
from .. import utils
from .. import config


def train(train_data_dict_path,
          val_data_dict_path,
          spect_params,
          networks,
          num_epochs,
          config_file,
          val_error_step=None,
          checkpoint_step=None,
          patience=None,
          save_only_single_checkpoint_file=True,
          normalize_spectrograms=False,
          root_results_dir=None,
          save_transformed_data=False,
          ):
    """train a single model using training set specified in config.ini file

    Parameters
    ----------
    train_data_dict_path : str
        path to training data
    val_data_dict_path : str
        path to validation data
    spect_params : dict
        parameters for creating spectrograms.
        Used to ensure that what's in config file matches what's in
        the data.
    networks : dict
        where each key is the name of a neural network and the corresponding
        value is the configuration for that network (in a namedtuple or a dict)
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    config_file : str
        path to config.ini file. Used to rewrite file with options determined by
        this function and needed for other functions (e.g. cli.summary)
    val_error_step : int
        step/epoch at which to estimate accuracy using validation set.
        Default is None, in which case no validation is done.
    checkpoint_step : int
        step/epoch at which to save to checkpoint file.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of epochs to wait without the error dropping before stopping the
        training. Default is None, in which case training continues for num_epochs
    save_only_single_checkpoint_file : bool
        if True, save only one checkpoint file instead of separate files every time
        we save. Default is True.
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    root_results_dir : str
        path in which to create results directory for this run of cli.learncurve
    save_transformed_data : bool
        if True, save transformed data (i.e. scaled, reshaped). The data can then
        be used on a subsequent run (e.g. if you want to compare results
        from different hyperparameters across the exact same training set).
        Also useful if you need to check what the data looks like when fed to networks.
        Default is False.

    Returns
    -------
    None

    Saves results in root_results_dir and adds some options to config_file.
    """
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    if root_results_dir:
        results_dirname = os.path.join(root_results_dir,
                                       'results_' + timenow)
    else:
        results_dirname = os.path.join('.', 'results_' + timenow)
    os.makedirs(results_dirname)
    # copy config file into results dir now that we've made the dir
    shutil.copy(config_file, results_dirname)

    logfile_name = os.path.join(results_dirname,
                                'logfile_from_train_' + timenow + '.log')
    logger = logging.getLogger(__name__)
    logger.setLevel('INFO')
    logger.addHandler(logging.FileHandler(logfile_name))
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.info('Logging results to {}'.format(results_dirname))
    logger.info('Using config file: {}'.format(config_file))

    logger.info('Loading training data from {}'.format(
        os.path.dirname(
            train_data_dict_path)))
    train_data_dict = joblib.load(train_data_dict_path)
    labels_mapping = train_data_dict['labels_mapping']
    if train_data_dict['spect_params'] == 'matlab':
        warnings.warn('Not checking parameters used to compute spectrogram in '
                      'training data,\n because spectrograms were created in '
                      'Matlab')
    else:
        if train_data_dict['spect_params'] != spect_params:
            raise ValueError('Spectrogram parameters in config file '
                             'do not match parameters specified in training data_dict.\n'
                             'Config file is: {}\n'
                             'Data dict is: {}.'.format(config_file,
                                                        train_data_dict_path))

    # save copy of labels_mapping in results directory
    labels_mapping_file = os.path.join(results_dirname, 'labels_mapping')
    with open(labels_mapping_file, 'wb') as labels_map_file_obj:
        pickle.dump(labels_mapping, labels_map_file_obj)

    # n_syllables, i.e., number of label classes to predict
    # Note that mapping includes label for silent gap b/t syllables
    # Error checking code to ensure that it is in fact a consecutive
    # series of integers from 0 to n, so we don't predict classes that
    # don't exist
    if sorted(labels_mapping.values()) != list(range(len(labels_mapping))):
        raise ValueError('Labels mapping does not map to a consecutive'
                         'series of integers from 0 to n (where 0 is the '
                         'silent gap label and n is the number of syllable'
                         'labels).')
    n_syllables = len(labels_mapping)
    logger.debug('n_syllables: '.format(n_syllables))

    # copy training data to results dir so we have it stored with results
    logger.info('copying {} to {}'.format(train_data_dict_path,
                                          results_dirname))
    shutil.copy(train_data_dict_path, results_dirname)

    (X_train,
     Y_train,
     X_train_spect_ID_vector,
     timebin_dur,
     files_used) = (train_data_dict['X_train'],
                    train_data_dict['Y_train'],
                    train_data_dict['spect_ID_vector'],
                    train_data_dict['timebin_dur'],
                    train_data_dict['filenames'])

    if Y_train.ndim > 1:
        # not clear to me right why labeled_timebins get saved as (n, 1)
        # instead of as (n) vector--i.e. if another functions depends on that shape
        # Below is hackish way around figuring that out.
        Y_train = np.squeeze(Y_train)

    logger.info('Size of each timebin in spectrogram, in seconds: {}'
                .format(timebin_dur))
    # dump filenames to a text file
    # to be consistent with what the matlab helper function does
    files_used_filename = os.path.join(results_dirname, 'training_filenames')
    with open(files_used_filename, 'w') as files_used_fileobj:
        files_used_fileobj.write('\n'.join(files_used))

    total_train_set_duration = X_train.shape[-1] * timebin_dur
    logger.info('Total duration of training set (in s): {}'
                .format(total_train_set_duration))

    # transpose X_train, so rows are timebins and columns are frequency bins
    # because cnn-bilstm network expects this orientation for input
    X_train = X_train.T

    val_data_dict = joblib.load(val_data_dict_path)
    (X_val,
     Y_val) = (val_data_dict['X_val'],
               val_data_dict['Y_val'])
    if val_data_dict['spect_params'] == 'matlab':
        warnings.warn('Not checking parameters used to compute spectrogram in '
                      'validation data,\n because spectrograms were created in '
                      'Matlab')
    else:
        if val_data_dict['spect_params'] != spect_params:
            raise ValueError('Spectrogram parameters in config file '
                             'do not match parameters specified in validation data_dict.\n'
                             'Config file is: {}\n'
                             'Data dict is: {}.'.format(config_file,
                                                        val_data_dict_path))
    #####################################################
    # note that we 'transpose' the spectrogram          #
    # so that rows are time and columns are frequencies #
    #####################################################
    X_val = X_val.T
    if save_transformed_data:
        joblib.dump(X_val, os.path.join(results_dirname, 'X_val'))
        joblib.dump(Y_val, os.path.join(results_dirname, 'Y_val'))

    logger.info('will measure error on validation set '
                'every {} steps of training'.format(val_error_step))
    logger.info('will save a checkpoint file '
                'every {} steps of training'.format(checkpoint_step))

    if save_only_single_checkpoint_file:
        logger.info('save_only_single_checkpoint_file = True\n'
                    'will save only one checkpoint file'
                    'and overwrite every {} steps of training'.format(checkpoint_step))
    else:
        logger.info('save_only_single_checkpoint_file = False\n'
                    'will save a separate checkpoint file '
                    'every {} steps of training'.format(checkpoint_step))

    logger.info('\'patience\' is set to: {}'.format(patience))

    logger.info('number of training epochs will be {}'
                .format(num_epochs))

    if normalize_spectrograms:
        logger.info('will normalize spectrograms')
        spect_scaler = utils.data.SpectScaler()
        X_train = spect_scaler.fit_transform(X_train)
        logger.info('normalizing validation set to match training set')
        X_val = spect_scaler.transform(X_val)
        joblib.dump(spect_scaler,
                    os.path.join(results_dirname, 'spect_scaler'))

    if save_transformed_data:
        scaled_data_filename = os.path.join(results_dirname,
                                            'scaled_spects')
        scaled_data_dict = {'X_train_scaled': X_train,
                            'X_val_scaled': X_val,
                            'Y_train_subset': Y_train}
        joblib.dump(scaled_data_dict, scaled_data_filename)

    freq_bins = X_train.shape[-1]  # number of columns
    logger.debug('freq_bins in spectrogram: '.format(freq_bins))

    NETWORKS = network._load()

    for net_name, net_config in networks.items():
        net_config_dict = net_config._asdict()
        net_config_dict['n_syllables'] = n_syllables
        if 'freq_bins' in net_config_dict:
            net_config_dict['freq_bins'] = freq_bins
        net = NETWORKS[net_name](**net_config_dict)

        results_dirname_this_net = os.path.join(results_dirname, net_name)

        checkpoint_filename = f'checkpoint_{net_name}'
        checkpoint_path = os.path.join(results_dirname_this_net,
                                       checkpoint_filename)

        if not os.path.isdir(results_dirname_this_net):
            os.makedirs(results_dirname_this_net)
        logs_subdir = f'log_{net_name}'
        logs_path = os.path.join(results_dirname_this_net,
                                 'logs',
                                 logs_subdir)
        if not os.path.isdir(logs_path):
            os.makedirs(logs_path)

        net.add_summary_writer(logs_path=logs_path)

        (X_val_batch,
         Y_val_batch,
         num_batches_val) = utils.data.reshape_data_for_batching(X_val,
                                                                 net_config.batch_size,
                                                                 net_config.time_bins,
                                                                 Y_val)

        if save_transformed_data:
            scaled_reshaped_data_filename = os.path.join(results_dirname_this_net,
                                                         'scaled_reshaped_spects')
            scaled_reshaped_data_dict = {'X_train_scaled_reshaped': X_train,
                                         'Y_train_reshaped': Y_train,
                                         'X_val_scaled_batch': X_val_batch,
                                         'Y_val_batch': Y_val_batch}
            joblib.dump(scaled_reshaped_data_dict, scaled_reshaped_data_filename)

        costs = []
        val_errs = []
        curr_min_err = 1  # i.e. 100%
        err_patience_counter = 0

        with tf.Session(graph=net.graph) as sess:
            sess.run(net.init)

            # figure out number of batches we can get out of subset of training data
            # if we slide a window along the spectrogram with a stride of 1
            # and use each window as one sample in a batch
            num_timebins_training_set = X_train.shape[0]
            num_windows = num_timebins_training_set - net_config.time_bins
            logger.info('training set with {} time bins will yield {} windows '
                        'of width {} time bins'
                        .format(num_timebins_training_set,
                                num_windows,
                                net_config.time_bins))
            num_batches = num_windows // net_config.batch_size  # note floor division
            # meaning we'll throw away some windows
            new_last_ind = net_config.batch_size * num_batches
            logger.info('divided into batches of size {} yields {} batches '
                        'for a total of {} windows (throwing away {}).'
                        .format(net_config.batch_size,
                                num_batches,
                                new_last_ind,
                                num_windows - new_last_ind))

            for epoch in range(num_epochs):
                # every epoch we are going to shuffle the order in which we look at every window
                shuffle_order = np.random.permutation(num_windows)
                shuffle_order = shuffle_order[:new_last_ind].reshape(num_batches, net_config.batch_size)
                pbar = tqdm(shuffle_order)
                for batch_num, batch_inds in enumerate(pbar):
                    X_batch = []
                    Y_batch = []
                    for start_ind in batch_inds:
                        X_batch.append(
                            X_train[start_ind:start_ind + net_config.time_bins, :]
                        )
                        Y_batch.append(
                            Y_train[start_ind:start_ind + net_config.time_bins]
                        )
                    X_batch = np.stack(X_batch)
                    Y_batch = np.stack(Y_batch)
                    d = {net.X: X_batch,
                         net.y: Y_batch,
                         net.lng: [net_config.time_bins] * net_config.batch_size}
                    _cost, _, summary = sess.run((net.cost,
                                                  net.optimize,
                                                  net.merged_summary_op),
                                                 feed_dict=d)
                    costs.append(_cost)
                    net.summary_writer.add_summary(summary, epoch)
                    pbar.set_description(
                        f"epoch {epoch + 1}, batch {batch_num + 1} of {num_batches}, cost: {_cost:8.4f}"
                    )

                if val_error_step:
                    if epoch % val_error_step == 0:
                        if 'Y_pred_val' in locals():
                            del Y_pred_val

                        for b in range(num_batches_val):  # "b" is "batch number"
                            X_b = X_val_batch[:, b * net_config.time_bins: (b + 1) * net_config.time_bins, :]
                            Y_b = Y_val_batch[:, b * net_config.time_bins: (b + 1) * net_config.time_bins]
                            d = {net.X: X_b,
                                 net.y: Y_b,
                                 net.lng: [net_config.time_bins] * net_config.batch_size}

                            if 'Y_pred_val' in locals():
                                preds = sess.run(net.predict, feed_dict=d)
                                preds = preds.reshape(net_config.batch_size, -1)
                                Y_pred_val = np.concatenate((Y_pred_val, preds), axis=1)
                            else:
                                Y_pred_val = sess.run(net.predict, feed_dict=d)
                                Y_pred_val = Y_pred_val.reshape(net_config.batch_size, -1)

                        # get rid of zero padding predictions
                        Y_pred_val = Y_pred_val.ravel()[:Y_val.shape[0], np.newaxis]
                        val_errs.append(np.sum(Y_pred_val - Y_val != 0) / Y_val.shape[0])
                        print("epoch {}, validation error: {}".format(epoch + 1, val_errs[-1]))

                    if patience:
                        if val_errs[-1] < curr_min_err:
                            # error went down, set as new min and reset counter
                            curr_min_err = val_errs[-1]
                            err_patience_counter = 0
                            print("Validation error improved.\n"
                                  "Saving checkpoint to {}".format(checkpoint_path))
                            net.saver.save(sess, checkpoint_path)
                        else:
                            err_patience_counter += 1
                            if err_patience_counter > patience:
                                print("stopping because validation error has not improved in {} epochs"
                                      .format(patience))
                                with open(os.path.join(results_dirname_this_net, "costs"), 'wb') as costs_file:
                                    pickle.dump(costs, costs_file)
                                with open(os.path.join(results_dirname_this_net, "val_errs"), 'wb') as val_errs_file:
                                    pickle.dump(val_errs, val_errs_file)
                                break

                if checkpoint_step:
                    if epoch % checkpoint_step == 0:
                        "Saving checkpoint."
                        if save_only_single_checkpoint_file is False:
                            checkpoint_path_tmp = checkpoint_path + '_{}'.format(epoch)
                        else:
                            checkpoint_path_tmp = checkpoint_path
                        net.saver.save(sess, checkpoint_path_tmp)
                        with open(os.path.join(results_dirname_this_net, "val_errs"), 'wb') as val_errs_file:
                            pickle.dump(val_errs, val_errs_file)

                if epoch == (num_epochs - 1):  # if this is the last epoch
                    "Reached max. number of epochs, saving checkpoint."
                    net.saver.save(sess, checkpoint_path)
                    with open(os.path.join(results_dirname_this_net, "costs"),
                              'wb') as costs_file:
                        pickle.dump(costs, costs_file)
                    with open(os.path.join(results_dirname_this_net, "val_errs"),
                              'wb') as val_errs_file:
                        pickle.dump(val_errs, val_errs_file)

    # lastly rewrite config file,
    # so that paths where results were saved are automatically in config
    config = ConfigParser()
    config.read(config_file)
    config.set(section='DATA',
               option='n_syllables',
               value=str(n_syllables))
    config.set(section='OUTPUT',
               option='results_dir_made_by_main_script',
               value=results_dirname)
    with open(config_file, 'w') as config_file_rewrite:
        config.write(config_file_rewrite)


if __name__ == "__main__":
    config_file = os.path.normpath(sys.argv[1])
    config = config.parse.parse_config(config_file)
    train(config_file)



"""
Python functions to facilitate interacting with the dataset from Koumura and
Okanoya 2016 [1].

data: https://figshare.com/articles/BirdsongRecognition/3470165

The original code was released under the GNU license:
https://github.com/cycentum/birdsong-recognition/blob/master/
birdsong-recognition/src/computation/ViterbiSequencer.java

[1] Koumura T, Okanoya K (2016) Automatic Recognition of Element Classes and
Boundaries in the Birdsong with Variable Sequences. PLoS ONE 11(7): e0159188.
doi:10.1371/journal.pone.0159188
"""

#from standard library
import os
import xml.etree.ElementTree as ET

#from dependencies
import numpy as np


class Syllable:
    """
    Object that represents a syllable.

    Properties
    ----------
    position : int
        starting sample number ("frame") within .wav file
        *** relative to start of sequence! ***
    length : int
        duration given as number of samples
    label : char
        text representation of syllable as classified by a human
        or a machine learning algorithm
    """
    def __init__(self, position, length, label):
        self.position = position
        self.length = length
        self.label = label

    def __repr__(self):
        rep_str = "Syllable labeled {} at position {} with length {}".format(
                   self.label,self.position,self.length)
        return rep_str


class Sequence:
    """
    Object that represents a sequence of syllables.

    Properties
    ----------
    wavFile : string
        file name of .wav file in which sequence occurs
    position : int
        starting sample number within .wav file
    length : int
        duration given as number of samples
    syls : list
        list of syllable objects that make up sequence
    seqSpect : spectrogram object
    """

    def __init__(self, wav_file,position, length, syl_list):
        self.wavFile = wav_file
        self.position = position
        self.length = length
        self.numSyls = len(syl_list)
        self.syls = syl_list
        self.seqSpect = None

    def __repr__(self):
        rep_str = "Sequence from {} with position {} and length {}".format(
                  self.wavFile,self.position,self.length)
        return rep_str


def parse_xml(xml_file,concat_seqs_into_songs=False):
    """
    parses Annotation.xml files from BirdsongRecognition dataset.

    Parameters
    ----------
    xml_file : string
        filename of .xml file, e.g. 'Annotation.xml'
    concat_seqs_into_songs : Boolean
        if True, concatenate sequences into songs, where each wav file is a
        song. Default is False.

    Returns
    -------
    seq_list : list of Sequence objects
    """

    tree = ET.ElementTree(file=xml_file)
    seq_list = []
    for seq in tree.iter(tag='Sequence'):
        wav_file = seq.find('WaveFileName').text
        position = int(seq.find('Position').text)
        length = int(seq.find('Length').text)
        syl_list = []
        for syl in seq.iter(tag='Note'):
            syl_position = int(syl.find('Position').text)
            syl_length = int(syl.find('Length').text)
            label = syl.find('Label').text

            syl_obj = Syllable(position = syl_position,
                               length = syl_length,
                               label = label)
            syl_list.append(syl_obj)
        seq_obj = Sequence(wav_file = wav_file,
                           position = position,
                           length = length,
                           syl_list = syl_list)
        seq_list.append(seq_obj)

    if concat_seqs_into_songs:
        song_list = []
        curr_wavFile = seq_list[0].wavFile
        new_seq_obj = seq_list[0]
        for syl in new_seq_obj.syls:
            syl.position += new_seq_obj.position

        for seq in seq_list[1:]:
            if seq.wavFile == curr_wavFile:
                new_seq_obj.length += seq.length
                new_seq_obj.numSyls += seq.numSyls
                for syl in seq.syls:
                    syl.position += seq.position
                new_seq_obj.syls += seq.syls

            else:
                song_list.append(new_seq_obj)
                curr_wavFile = seq.wavFile
                new_seq_obj = seq
                for syl in new_seq_obj.syls:
                    syl.position += new_seq_obj.position

        song_list.append(new_seq_obj)  # to append last song

        return song_list

    else:
        return seq_list


def load_song_annot(songfiles, annotation_file):
    """

    Parameters
    ----------
    songfiles: list
        of str, filenames of .wav files from BirdsongRecogntion dataset
    annotation_file: str
        path to Annotation.xml file

    Returns
    -------
    all_songfile_dict: dict
        with keys onsets, offsets, and labels
    """
    seq_list = parse_xml(annotation_file,
                         concat_seqs_into_songs=True)
    wav_files = [seq.wavFile for seq in seq_list]

    annotation_dict = {}
    for songfile in songfiles:
        songfile = os.path.basename(songfile)
        ind = wav_files.index(songfile)
        this_seq = seq_list[ind]
        onsets = np.asarray([syl.position for syl in this_seq.syls])
        offsets = np.asarray([syl.position + syl.length for syl in this_seq.syls])
        labels = [syl.label for syl in this_seq.syls]
        songfile_dict = {
            'onsets' : onsets,
            'offsets' : offsets,
            'labels' : labels
        }
        annotation_dict[songfile] = songfile_dict
    return annotation_dict


def determine_unique_labels(annotation_file):
    """given an annotation.xml file
    from a bird in BirdsongRecognition dataset,
    determine unique set of labels applied to syllables from that bird"""
    annotation = parse_xml(annotation_file,
                           concat_seqs_into_songs=True)
    lbls = [syl.label
            for seq in annotation
            for syl in seq.syls]
    unique_lbls = np.unique(lbls).tolist()
    unique_lbls = ''.join(unique_lbls)  # convert from list to string
    return unique_lbls



import os
from glob import glob

from scipy.io import loadmat
import joblib
import numpy as np

from vak.utils.data import vec_translate


def convert_mat_to_spect(mat_spect_files,
                         mat_spects_annotation_file,
                         output_dir,
                         labels_mapping=None,
                         skip_files_with_labels_not_in_labelset=True,
                         n_decimals_trunc=3):
    """converts .mat files with spectrograms to .spect files
    that are used by make_data.py script and the make_data_dicts
    function that it calls.

    Parameters
    ----------
    mat_spect_files : list
        of str, full path to .mat files
        path to directory containing .mat files
    mat_spects_annotation_file : str
        full path to annotation file containing 'keys' and 'elements'
        where 'keys' are filenames of audio files and 'elements'
        contains additional annotation not found in .mat files
    output_dir : str
        full path to directory where .spect files will be saved
    labels_mapping : dict
        maps str labels to consecutive integer values {0,1,2,...N} where N
        is the number of classes / label types.
        Default is None -- currently not implemented.
    skip_files_with_labels_not_in_labelset : bool
        if True, skip .mat files where the 'labels' contains str labels
        not found in labels_mapping
    n_decimals_trunc : int
        number of decimal places to keep when truncating timebin_dur
        default is 3

    Returns
    -------
    spect_files_path : str
        Full path to file called 'spect_files'
        which contains a list of three-element tuples:
            spect_filename : str, filename of `.spect` file
            spect_dur : float, duration of the spectrogram from cbin
            labels : str, labels from .cbin.not.mat associated with .cbin
                     (string labels for syllables in spectrogram)
        Used when building data sets of a specific duration.

    The fucntion saves a .spect file for each .mat file containing a
    spectrogram. Each .mat file with a spectrogram should contains the following
    keys:
        s : ndarray
            the spectrogram
        f : ndarray
            vector of values at bin centers for frequencies in spectrogram
        t : ndarray
            vector of values for bin centers for times in spectrogram
        labels : ndarray
            vector of same length as t where each value is a label for that time bin
    containing the spectrograms.

    If a .mat file does not contain these keys, the function skips that file.

    """
    decade = 10**n_decimals_trunc  # used below for truncating floating point

    annotations = loadmat(mat_spects_annotation_file, squeeze_me=True)
    annotations = dict(zip(annotations['keys'],
                          annotations['elements']))
    spect_files = []
    num_spect_files = len(mat_spect_files)

    # need to map values in 'labels' variable of .mat file
    # (which is a label for every time bin)
    # **from** int to **consecutive** ints
    # so, need to convert str keys in labels_mapping to int keys.
    # This is used in for loop below, with "vec_translate".
    int_labels_mapping = {}
    for key, val in labels_mapping.items():
        try:
            int_labels_mapping[int(key)] = val
        except ValueError:
            if key == 'silent_gap_label':
                int_labels_mapping[labels_mapping['silent_gap_label']] = \
                    labels_mapping['silent_gap_label']
            else:
                raise ValueError('could not convert key {} '
                                 'to int value for use with labeled '
                                 'time bin vectors in .mat files'
                                 .format(key))

    for filenum, matspect_filename in enumerate(mat_spect_files):
        print('loading annotation info from {}, file {} of {}'
              .format(matspect_filename, filenum, num_spect_files))
        wav_filename = os.path.basename(matspect_filename).replace('.mat',
                                                                '.wav')
        annotation = annotations[wav_filename]
        # below, the first .tolist() does not actually create list,
        # instead gets ndarray out of a zero-length ndarray of dtype=object.
        # This is just weirdness that results from loading complicated data
        # structure in .mat file.
        labels = annotation['segType'].tolist()
        if type(labels) == int:
            # this happens when there's only one syllable in the file
            # with only one corresponding label
            labels = [labels]  # so make it a one-element list
        elif type(labels) == np.ndarray:
            # second .tolist() then converts ndarray into list of ints
            labels = labels.tolist()
        else:
            raise ValueError("Unable to load labels from {}, because "
                             "the segType parsed as type {} which is "
                             "not recognized.".format(wav_filename,
                                                      type(labels)))

        if skip_files_with_labels_not_in_labelset:
            labels_set = set(labels)
            # below, set(labels_mapping) is a set of that dict's keys
            if not labels_set.issubset(set(labels_mapping)):
                extra_labels = labels_set - set(labels_mapping)
                # because there's some label in labels
                # that's not in labels_mapping
                print('Found labels, {}, in {}, that are not in labels_mapping.'
                      'Skipping file.'.format(extra_labels,
                                             matspect_filename))
                continue

        if all([type(label)==int for label in labels]):
            # to make consistent with .cbin and .wav files,
            # makes error checking in make_dicts function properly
            labels = [str(label) for label in labels]
        else:
            # currently function expects int labels
            raise TypeError('type of labels in {} not recognized, '
                            'must be all of type int'
                            .format(matspect_filename))

        matspect = loadmat(matspect_filename, squeeze_me=True)
        if 's' not in matspect:
            print('Did not find a spectrogram in {}. '
                  'Skipping this file.'.format(matspect_filename))
            continue

        if 'freq_bins' not in locals() and 'time_bins' not in locals():
            freq_bins = matspect['f']
            time_bins = matspect['t']
            timebin_dur = np.around(np.mean(np.diff(time_bins)), decimals=3)
            # below truncates any decimal place past decade
            timebin_dur = np.trunc(timebin_dur * decade) / decade
        else:
            if not np.array_equal(matspect['f'], freq_bins):
                raise ValueError('freq_bins in {} does not freq_bins from '
                                 'other .mat files'.format(matspect_filename))
            curr_file_timebin_dur = np.around(np.mean(np.diff(matspect['t'])),
                                              decimals=3)
            # below truncates any decimal place past decade
            curr_file_timebin_dur = np.trunc(curr_file_timebin_dur
                                             * decade) / decade
            if not np.allclose(curr_file_timebin_dur, timebin_dur):
                raise ValueError('duration of timebin in file {} did not '
                                 'match duration of timebin from other .mat '
                                 'files.'.format(matspect_filename))

        # number of freq. bins should equal number of rows
        if matspect['f'].shape[-1] != matspect['s'].shape[0]:
            raise ValueError('length of freq_bins in {} does not match '
                             'number of rows in spectrogram'
                             .format(matspect_filename))
        # number of time bins should equal number of columns
        if matspect['t'].shape[-1] != matspect['s'].shape[1]:
            raise ValueError('length of time_bins in {} does not match '
                             'number of columns in spectrogram'
                             .format(matspect_filename))
        labeled_timebins = matspect['labels']
        if labeled_timebins.shape[-1] != matspect['t'].shape[-1]:
            raise ValueError("length of 'labels' (labeled timebins vector)"
                             " in {} does not match number of time bins"
                             .format(matspect_filename))
        if labeled_timebins.ndim != 2:
            if labeled_timebins.ndim == 1:
                # make same shape as output of utils.make_labeled_timebins
                # so learn_curve.py doesn't crash when concatenating
                # with zero-pad vector
                labeled_timebins = labeled_timebins[:, np.newaxis]
            else:
                raise ValueError('labeled timebins vector from {} has '
                                 'invalid number of dimensions: {}'
                                 .format(matspect_filename, labels.ndim))

        labeled_timebins = vec_translate(labeled_timebins,
                                         int_labels_mapping)

        spect_dict = {'spect': matspect['s'],
                      'freq_bins': matspect['f'],
                      'time_bins': matspect['t'],
                      'labels': labels,
                      'labeled_timebins': labeled_timebins,
                      'timebin_dur': timebin_dur,
                      'spect_params': 'matlab',
                      'labels_mapping': labels_mapping}

        spect_filename = os.path.join(
            os.path.normpath(output_dir),
            os.path.basename(matspect_filename) + '.spect')
        joblib.dump(spect_dict, spect_filename)

        spect_dur = matspect['s'].shape[-1] * timebin_dur
        spect_files.append((spect_filename, spect_dur, labels))

    spect_files_path = os.path.join(output_dir, 'spect_files')
    joblib.dump(spect_files, spect_files_path)

    return spect_files_path


def convert_train_keys_to_txt(train_keys_path,
                              txt_filename = 'spect_files'):
    """get train_keys cell array out of .mat file, convert to list of str, save as .txt

    Parameters
    ----------
    train_keys_path : str
        path to folder with train_keys.mat file
    txt_filename : str
        filename for .txt file that contains list of .mat filenames
        Default is `training_filenames`

    Returns
    -------
    None. Saves .txt file in train_keys_path.
    """
    train_spect_files = loadmat(train_keys_path,
                                squeeze_me=True)['train_keys'].tolist()
    txt_filename = os.path.join(os.path.split(train_keys_path)[0],
                                txt_filename)
    with open(txt_filename, 'w') as fileobj:
        fileobj.write('\n'.join(train_spect_files))


import copy
import itertools
import os
import random
from glob import glob

import joblib
import numpy as np
from scipy.io import wavfile, loadmat
from tqdm import tqdm

from . import spect as spect_utils  # so as not to confuse with variable name `spect`
from vak import evfuncs
from vak.koumura_utils import load_song_annot


def vec_translate(a, my_dict):
    return np.vectorize(my_dict.__getitem__)(a)


# adapted from:
# https://github.com/NickleDave/hybrid-vocal-classifier/blob/master/hvc/neuralnet/utils.py
class SpectScaler:
    """class that scales spectrograms that all have the
    same number of frequency bins. Any input spectrogram
    will be scaled by subtracting off the mean of each
    frequency bin from the 'fit' set of spectrograms, and
    then dividing by the standard deviation of each
    frequency bin from the 'fit' set.
    """

    def __init__(self):
        pass

    def fit(self, spect):
        """fit a SpectScaler.
        Input should be spectrogram,
        oriented so that the columns are frequency bins.
        Fit function finds the mean and standard deviation of
        each frequency bin, which are used by `transform` method
        to scale other spectrograms.

        Parameters
        ----------
        spect : 2-d numpy array
            with dimensions (time bins, frequency bins)
        """

        if spect.ndim != 2:
            raise ValueError('input spectrogram should be a 2-d array')

        self.columnMeans = np.mean(spect, axis=0)
        self.columnStds = np.std(spect, axis=0)
        assert self.columnMeans.shape[-1] == spect.shape[-1]
        assert self.columnStds.shape[-1] == spect.shape[-1]
        self.nonZeroStd = np.argwhere(self.columnStds != 0)

    def _transform(self, spect):
        """transforms input spectrogram by subtracting off fit mean
        and then dividing by standard deviation
        """

        transformed = spect - self.columnMeans
        # to keep any zero stds from causing NaNs
        transformed[:, self.nonZeroStd] = (
            transformed[:, self.nonZeroStd] / self.columnStds[self.nonZeroStd])
        return transformed

    def transform(self, spects):
        """normalizes input spectrograms with fit parameters
        Assumes spectrograms are oriented with columns being frequency bins
        and rows being time bins.

        Parameters
        ----------
        spects : 2-d numpy array or list of 2-d numpy arrays
            with dimensions (time bins, frequency bins)

        """

        if any([not hasattr(self, attr) for attr in ['columnMeans',
                                                     'columnStds']]):
            raise AttributeError('SpectScaler properties are set to None,'
                                 'must call fit method first to set the'
                                 'value of these properties before calling'
                                 'transform')

        if type(spects) != np.ndarray and type(spects) != list:
            raise TypeError('type {} is not valid for spects'
                            .format(type(spects)))

        if type(spects) == np.ndarray:
            if spects.shape[-1] != self.columnMeans.shape[-1]:
                raise ValueError('number of columns in spects, {}, '
                                 'does not match shape of self.columnMeans, {},'
                                 'i.e. the number of columns from the spectrogram'
                                 'to which the scaler was fit originally')
            return self._transform(spects)

        elif type(spects) == list:
            z_norm_spects = []
            for spect in spects:
                z_norm_spects.append(self._transform(spect))

            return z_norm_spects

    def fit_transform(self, spects):
        """first calls fit and then returns normalized spects
        transformed using the fit parameters"""

        if type(spects) != np.ndarray:
            raise TypeError('spects passed to fit_transform '
                            'should be numpy array, not {}'
                            .format(type(spects)))

        if spects.ndim != 2:
            raise ValueError('ndims of spects should be 2, not {}'
                             .format(spects.ndim))

        self.fit(spects)
        return self.transform(spects)


def make_labeled_timebins_vector(labels,
                                 onsets,
                                 offsets,
                                 time_bins,
                                 silent_gap_label=0):
    """makes a vector of labels for each timebin from a spectrogram,
    given labels for syllables plus onsets and offsets of syllables

    Parameters
    ----------
    labels : ints
        should be mapping returned by make_labels_mapping
    onsets : ndarray
        1d vector of floats, syllable onsets in seconds
    offsets : ndarray
        1d vector of floats, offsets in seconds
    time_bins : ndarray
        1d vector of floats,
        time in seconds for each time bin of a spectrogram
    silent_gap_label : int
        label assigned to silent gaps
        default is 0

    Returns
    -------
    label_vec : ndarray
        same length as time_bins, with each element a label for
        each time bin
    """

    labels = [int(label) for label in labels]
    label_vec = np.ones((time_bins.shape[-1], 1), dtype='int8') * silent_gap_label
    onset_inds = [np.argmin(np.abs(time_bins - onset))
                  for onset in onsets]
    offset_inds = [np.argmin(np.abs(time_bins - offset))
                   for offset in offsets]
    for label, onset, offset in zip(labels, onset_inds, offset_inds):
        label_vec[onset:offset+1] = label
        # offset_inds[ind]+1 because of Matlab one-indexing
    return label_vec


def make_spects_from_list_of_files(filelist,
                                   spect_params,
                                   output_dir,
                                   labels_mapping,
                                   skip_files_with_labels_not_in_labelset=True,
                                   annotation_file=None,
                                   n_decimals_trunc=3,
                                   is_for_predict=False):
    """makes spectrograms from a list of audio files

    Parameters
    ----------
    filelist : list
        of str, full paths to .wav or .cbin files
        from which to make spectrograms
    spect_params : dict
        parameters for computing spectrogram, from .ini file
    output_dir : str
        directory in which to save .spect file generated for each .cbin file,
        as described below
    labels_mapping : dict
        maps str labels to consecutive integer values {0,1,2,...N} where N
        is the number of classes / label types
    skip_files_with_labels_not_in_labelset : bool
        if True, skip .cbin files where the 'labels' array in the corresponding
        .cbin.not.mat file contains str labels not found in labels_mapping
    annotation_file : str
        full path to file with annotations. Required for .wav files. Expected to
        be a .mat file containing the variables 'keys' and 'elements', where 'keys'
        is a list of .wav filenames and 'elements' are the associated annotations
        for each filename.
        Default is None.
    n_decimals_trunc : int
        number of decimal places to keep when truncating timebin_dur
        default is 3
    is_for_predict : bool
        if True, we are making spectrograms for prediction of labels. This tells function not
        to worry about annotation and to just insert dummy vector of labeled timebins
        into .spect files. Default is False.

    Returns
    -------
    spects_used_path : str
        Full path to file called 'spect_files'
        which contains a list of three-element tuples:
            spect_filename : str, filename of `.spect` file
            spect_dur : float, duration of the spectrogram from cbin
            labels : str, labels from .cbin.not.mat associated with .cbin
                     (string labels for syllables in spectrogram)
        Used when building data sets of a specific duration.

    For each .wav or .cbin filename in the list, a '.spect' file is saved.
    Each '.spect' file contains a "pickled" Python dictionary
    with the following key, value pairs:
        spect : ndarray
            spectrogram
        freq_bins : ndarray
            vector of centers of frequency bins from spectrogram
        time_bins : ndarray
            vector of centers of tme bins from spectrogram
        labeled_timebins : ndarray
            same length as time_bins, but value of each element is a label
            corresponding to that time bin
    """
    decade = 10** n_decimals_trunc  # used below for truncating floating point

    if all(['.cbin' in filename for filename in filelist]):
        filetype = 'cbin'
    elif all(['.wav' in filename for filename in filelist]):
        filetype = 'wav'
    else:
        raise ValueError('Could not determine whether filelist is '
                         'a list of .cbin files or of .wav files. '
                         'All files must be of the same type.')

    if filetype == 'wav':
        if annotation_file is None and is_for_predict is False:
            raise ValueError('annotation_file is required when using .wav files')
        elif annotation_file and is_for_predict is True:
            if annotation_file.endswith('.mat'):
                # the complicated nested structure of the annotation.mat files
                # (a cell array of Matlab structs)
                # makes it hard for loadmat to load them into numpy arrays.
                # Some of the weird looking things below, like accessing fields and
                # then converting them to lists, are work-arounds to deal with
                # the result of loading the complicated structure.
                # Setting squeeze_me=True gets rid of some but not all of the weirdness.
                annotations = loadmat(annotation_file, squeeze_me=True)
                # 'keys' here refers to filenames, which are 'keys' for the 'elements'
                keys_key = [key for key in annotations.keys() if 'keys' in key]
                elements_key = [key for key in annotations.keys() if 'elements' in key]
                if len(keys_key) > 1:
                    raise ValueError('Found more than one `keys` in annotations.mat file')
                if len(elements_key) > 1:
                    raise ValueError('Found more than one `elements` in annotations.mat file')
                if len(keys_key) < 1:
                    raise ValueError('Did not find `keys` in annotations.mat file')
                if len(elements_key) < 1:
                    raise ValueError('Did not find `elements` in annotations.mat file')
                keys_key = keys_key[0]
                elements_key = elements_key[0]
                annot_keys = annotations[keys_key].tolist()
                annot_elements = annotations[elements_key]
            elif annotation_file.endswith('.xml'):
                annotation_dict = load_song_annot(filelist, annotation_file)
        elif annotation_file is None and is_for_predict is True:
            pass
        else:
            ValueError('make_spects_from_list_of_files received annotation file '
                       'but is_for_predict is True')

    # need to keep track of name of files used since we may skip some.
    # (cbins_used is actually a list of tuples as defined in docstring)
    spect_files = []

    pbar = tqdm(filelist)
    for filename in pbar:
        basename = os.path.basename(filename)
        if filetype == 'cbin':
            if not is_for_predict:
                try:
                    notmat_dict = evfuncs.load_notmat(filename)
                except FileNotFoundError:
                    pbar.set_description(
                        f'Did not find .not.mat file for {basename}, skipping file.'
                    )
                    continue
                this_labels_str = notmat_dict['labels']
                onsets = notmat_dict['onsets'] / 1000
                offsets = notmat_dict['offsets'] / 1000
            dat, fs = evfuncs.load_cbin(filename)

        elif filetype == 'wav':
            fs, dat = wavfile.read(filename)
            if not is_for_predict:
                if annotation_file.endswith('.mat'):
                    ind = annot_keys.index(os.path.basename(filename))
                    annotation = annot_elements[ind]
                    # The .tolist() methods calls below are to get the
                    # array out of the weird lengthless object array
                    # that scipy.io.loadmat produces when trying to load
                    # the annotation files.
                    this_labels_str = annotation['segType'].tolist()
                    onsets = annotation['segFileStartTimes'].tolist()
                    offsets = annotation['segFileEndTimes'].tolist()
                elif annotation_file.endswith('.xml'):
                    filename_key = os.path.basename(filename)
                    this_labels_str = annotation_dict[filename_key]['labels']
                    onsets = annotation_dict[filename_key]['onsets'] / fs
                    offsets = annotation_dict[filename_key]['offsets'] / fs

        if not is_for_predict:
            if skip_files_with_labels_not_in_labelset:
                labels_set = set(this_labels_str)
                # below, set(labels_mapping) is a set of that dict's keys
                if not labels_set.issubset(set(labels_mapping)):
                    # because there's some label in labels
                    # that's not in labels_mapping
                    pbar.set_description(
                        f'found labels in {basename} not in labels_mapping, skipping file'
                    )
                    continue

        pbar.set_description(f'making .spect file for {basename}')

        if 'freq_cutoffs' in spect_params:
            dat = spect_utils.butter_bandpass_filter(dat,
                                                     spect_params['freq_cutoffs'][0],
                                                     spect_params['freq_cutoffs'][1],
                                                     fs)

        # have to make a new dictionary with args to spectrogram
        # instead of e.g. using spect_params._asdict()
        # because we don't want spect_params.freq_cutoffs as an arg to spect_utils.spectrogram
        specgram_params = {'fft_size': spect_params.fft_size,
                           'step_size': spect_params.step_size}
        if 'thresh' in spect_params:
            specgram_params['thresh'] = spect_params.thresh
        if 'transform_type' in spect_params:
            specgram_params['transform_type'] = spect_params.transform_type

        spect, freq_bins, time_bins = spect_utils.spectrogram(dat, fs,
                                                              **specgram_params)

        if 'freq_cutoffs' in spect_params:
            f_inds = np.nonzero((freq_bins >= spect_params['freq_cutoffs'][0]) &
                                (freq_bins < spect_params['freq_cutoffs'][1]))[0]  # returns tuple
            spect = spect[f_inds, :]
            freq_bins = freq_bins[f_inds]

        if not is_for_predict:
            this_labels = [labels_mapping[label]
                           for label in this_labels_str]
            this_labeled_timebins = make_labeled_timebins_vector(this_labels,
                                                                 onsets,
                                                                 offsets,
                                                                 time_bins,
                                                                 labels_mapping['silent_gap_label'])
        elif is_for_predict:
            this_labels = []
            this_labels_str = ''
            # below 1 because rows = freq bins, cols = time_bins before we take transpose (as we do in main loop)
            this_labeled_timebins = np.zeros((spect.shape[1], 1))

        if not 'timebin_dur' in locals():
            timebin_dur = np.around(np.mean(np.diff(time_bins)), decimals=3)
        else:
            curr_timebin_dur = np.around(np.mean(np.diff(time_bins)), decimals=3)
            # below truncates any decimal place past decade
            curr_timebin_dur = np.trunc(curr_timebin_dur
                                        * decade) / decade
            if not np.allclose(curr_timebin_dur, timebin_dur):
                raise ValueError("duration of timebin in file {}, {}, did not "
                                 "match duration of timebin from other .mat "
                                 "files, {}.".format(curr_timebin_dur,
                                                     filename,
                                                     timebin_dur))
        spect_dur = time_bins.shape[-1] * timebin_dur

        spect_dict = {'spect': spect,
                      'freq_bins': freq_bins,
                      'time_bins': time_bins,
                      'labels': this_labels_str,
                      'labeled_timebins': this_labeled_timebins,
                      'timebin_dur': timebin_dur,
                      'spect_params': spect_params,
                      'labels_mapping': labels_mapping}

        spect_dict_filename = os.path.join(
            os.path.normpath(output_dir),
            os.path.basename(filename) + '.spect')
        joblib.dump(spect_dict, spect_dict_filename)

        spect_files.append((spect_dict_filename, spect_dur, this_labels_str))

    spect_files_path = os.path.join(output_dir, 'spect_files')
    joblib.dump(spect_files, spect_files_path)

    return spect_files_path


def make_data_dicts(output_dir,
                    total_train_set_duration,
                    validation_set_duration,
                    test_set_duration,
                    labelset,
                    spect_files=None):
    """function that loads data and saves in dictionaries

    Parameters
    ----------
    output_dir : str
        path to output_dir containing .spect files
    total_train_set_duration : float
    validation_set_duration : float
    test_set_duration : float
        all in seconds
    labelset : list
        of str, labels used
    spect_files : str
        full path to file containing 'spect_files' list of tuples
        saved by function make_spects_from_list_of_files.
        Default is None, in which case this function looks for
        a file named 'spect_files' in output_dir.

    Returns
    -------
    saved_data_dict_paths : dict
        with keys {'train', 'test', 'val'} and values being the path
        to which the data_dict was saved

    Saves three 'data_dict' files (train, validation, and test)
    in output_dir, with following structure:
        spects : list
            of ndarray, spectrograms from audio files
        filenames : list
            same length as spects, filename of each audio file that was converted to spectrogram
        freq_bins : ndarray
            vector of frequencies where each value is a bin center. Same for all spectrograms
        time_bins : list
            of ndarrays, each a vector of times where each value is a bin center.
            One for each spectrogram
        labelset : list
            of strings, labels corresponding to each spectrogram
        labeled_timebins : list
            of ndarrays, each same length as time_bins but value is a label for that bin.
            In other words, the labels vector is mapped onto the time_bins vector for the
            spectrogram.
        X : ndarray
            X_train, X_val, or X_test, depending on which data_dict you are looking at.
            Some number of spectrograms concatenated, enough so that the total duration
            of the spectrogram in time bins is equal to or greater than the target duration.
            If greater than target, then X is truncated so it is equal to the target.
        Y : ndarray
            Concatenated labeled_timebins vectors corresponding to the spectrograms in X.
        spect_ID_vector : ndarray
            Vector where each element is an ID for a song. Used to randomly grab subsets
            of data of a target duration while still having the subset be composed of
            individual songs as much as possible. So this vector will look like:
            [0, 0, 0, ..., 1, 1, 1, ... , n, n, n] where n is equal to or (a little) less
            than the length of spects. spect_ID_vector.shape[-1] is the same as X.shape[-1]
            and Y.shape[0].
        timebin_dur : float
            duration of a timebin in seconds from spectrograms
        spect_params : dict
            parameters for computing spectrogram as specified in config.ini file.
            Will be checked against .ini file when running other cli such as learn_curve.py
        labels_mapping : dict
            maps str labels for syllables to consecutive integers.
            As explained in docstring for make_spects_from_list_of_files.
    """
    if not os.path.isdir(output_dir):
        raise NotADirectoryError('{} not recognized '
                                 'as a directory'.format(output_dir))

    if spect_files is None:
        spect_files = glob(os.path.join(output_dir,'spect_files'))
        if spect_files == []:  # if glob didn't find anything
            raise FileNotFoundError("did not find 'spect_files' file in {}"
                                    .format(output_dir))
        elif len(spect_files) > 1:
            raise ValueError("found than more than one 'spect_files' in {}:\n{}"
                             .format(output_dir, spect_files))
        else:
            spect_files = spect_files[0]

    if not os.path.isfile(spect_files):
        raise FileNotFoundError('{} not recognized as a file'
                                .format(spect_files))

    spect_files = joblib.load(spect_files)

    total_spects_dur = sum([spect[1] for spect in spect_files])
    total_dataset_dur = sum([total_train_set_duration,
                             validation_set_duration,
                             test_set_duration])
    if total_spects_dur < total_dataset_dur:
        raise ValueError('Total duration of all .cbin files, {} seconds,'
                         ' is less than total target duration of '
                         'training, validation, and test sets, '
                         '{} seconds'
                         .format(total_spects_dur, total_dataset_dur))

    # main loop that gets datasets
    iter = 1
    all_labels_err = ('Did not successfully divide data into training, '
                      'validation, and test sets of sufficient duration '
                      'after 1000 iterations.'
                      ' Try increasing the total size of the data set.')

    while 1:
        spect_files_copy = copy.deepcopy(spect_files)

        train_spects = []
        val_spects = []
        test_spects = []

        total_train_dur = 0
        val_dur = 0
        test_dur = 0

        choice = ['train', 'val', 'test']

        while 1:
            # pop tuples off cbins_used list and append to randomly-chosen
            # list, either train, val, or test set.
            # Do this until the total duration for each data set is equal
            # to or greater than the target duration for each set.
            try:
                ind = random.randint(0, len(spect_files_copy)-1)
            except ValueError:
                if len(spect_files_copy) == 0:
                    print('Ran out of spectrograms while dividing data into training, '
                          'validation, and test sets of specified durations. Iteration {}'
                          .format(iter))
                    iter += 1
                    break  # do next iteration
                else:
                    raise
            a_spect = spect_files_copy.pop(ind)
            which_set = random.randint(0, len(choice)-1)
            which_set = choice[which_set]
            if which_set == 'train':
                train_spects.append(a_spect)
                total_train_dur += a_spect[1]  # ind 1 is duration
                if total_train_dur >= total_train_set_duration:
                    choice.pop(choice.index('train'))
            elif which_set == 'val':
                val_spects.append(a_spect)
                val_dur += a_spect[1]  # ind 1 is duration
                if val_dur >= validation_set_duration:
                    choice.pop(choice.index('val'))
            elif which_set == 'test':
                test_spects.append(a_spect)
                test_dur += a_spect[1]  # ind 1 is duration
                if test_dur >= test_set_duration:
                    choice.pop(choice.index('test'))

            if len(choice) < 1:
                if np.sum(total_train_dur +
                                  val_dur +
                                  test_dur) < total_dataset_dur:
                    raise ValueError('Loop to find subsets completed but '
                                     'total duration of subsets is less than '
                                     'total duration specified by config file.')
                else:
                    break

            if iter > 1000:
                raise ValueError('Could not find subsets of sufficient duration in '
                                 'less than 1000 iterations.')

        # make sure no contamination between data sets.
        # If this is true, each set of filenames should be disjoint from others
        train_spect_files = [tup[0] for tup in train_spects]  # tup = a tuple
        val_spect_files = [tup[0] for tup in val_spects]
        test_spect_files = [tup[0] for tup in test_spects]
        assert set(train_spect_files).isdisjoint(val_spect_files)
        assert set(train_spect_files).isdisjoint(test_spect_files)
        assert set(val_spect_files).isdisjoint(test_spect_files)

        # make sure that each set contains all classes we
        # want the network to learn
        train_labels = itertools.chain.from_iterable(
            [spect[2] for spect in train_spects])
        train_labels = set(train_labels)  # make set to get unique values

        val_labels = itertools.chain.from_iterable(
            [spect[2] for spect in val_spects])
        val_labels = set(val_labels)

        test_labels = itertools.chain.from_iterable(
            [spect[2] for spect in test_spects])
        test_labels = set(test_labels)

        if train_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                print('Train labels did not contain all labels in labelset. '
                      'Getting new training set. Iteration {}'
                      .format(iter))
                continue
        elif val_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                print('Validation labels did not contain all labels in labelset. '
                      'Getting new validation set. Iteration {}'
                      .format(iter))
                continue
        elif test_labels != set(labelset):
            iter += 1
            if iter > 1000:
                raise ValueError(all_labels_err)
            else:
                print('Test labels did not contain all labels in labelset. '
                      'Getting new test set. Iteration {}'
                      .format(iter))
                continue
        else:
            break

    saved_data_dict_paths = {}

    for dict_name, spect_list, target_dur in zip(['train','val','test'],
                                                 [train_spects,val_spects,test_spects],
                                                 [total_train_set_duration,
                                                  validation_set_duration,
                                                  test_set_duration]):

        spects = []
        filenames = []
        all_time_bins = []
        labels = []
        labeled_timebins = []
        spect_ID_vector = []

        for spect_ind, spect_file in enumerate(spect_list):
            spect_dict = joblib.load(spect_file[0])
            spects.append(spect_dict['spect'])
            filenames.append(spect_file[0])
            all_time_bins.append(spect_dict['time_bins'])
            labels.append(spect_dict['labels'])
            labeled_timebins.append(spect_dict['labeled_timebins'])
            spect_ID_vector.extend([spect_ind] * spect_dict['time_bins'].shape[-1])

            if 'freq_bins' in locals():
                assert np.array_equal(spect_dict['freq_bins'], freq_bins)
            else:
                freq_bins = spect_dict['freq_bins']

            if 'labels_mapping' in locals():
                assert spect_dict['labels_mapping'] == labels_mapping
            else:
                labels_mapping = spect_dict['labels_mapping']

            if 'timebin_dur' in locals():
                assert spect_dict['timebin_dur'] == timebin_dur
            else:
                timebin_dur = spect_dict['timebin_dur']

            if 'spect_params' in locals():
                assert spect_dict['spect_params'] == spect_params
            else:
                spect_params = spect_dict['spect_params']

        X = np.concatenate(spects, axis=1)
        Y = np.concatenate(labeled_timebins)
        spect_ID_vector = np.asarray(spect_ID_vector, dtype='int')
        assert X.shape[-1] == Y.shape[0]  # Y has shape (timebins, 1)
        if X.shape[-1] > target_dur / timebin_dur:
            correct_length = np.round(target_dur / timebin_dur).astype(int)
            X = X[:, :correct_length]
            Y = Y[:correct_length, :]
            spect_ID_vector = spect_ID_vector[:correct_length]

        data_dict = {'spects': spects,
                     'filenames': filenames,
                     'freq_bins': freq_bins,
                     'time_bins': all_time_bins,
                     'labels': labels,
                     'labeled_timebins': labeled_timebins,
                     'spect_ID_vector': spect_ID_vector,
                     'X_' + dict_name: X,
                     'Y_' + dict_name: Y,
                     'timebin_dur': timebin_dur,
                     'spect_params': spect_params,
                     'labels_mapping': labels_mapping}

        print('saving data dictionary in {}'.format(output_dir))
        data_dict_path = os.path.join(output_dir, dict_name + '_data_dict')
        joblib.dump(data_dict, data_dict_path)
        saved_data_dict_paths[dict_name] = data_dict_path

    return saved_data_dict_paths


def make_data_dict_from_spect_files(labelset, spect_files='./spect_files',
                                    output_dir='.', dict_name='test'):
    """load data from a list of spect files and save into a dictionary.
    For running a summary of prediction error on a list of files you choose.

    Parameters
    ----------
    labelset : list
        of str, labels used
    spect_files : str
        full path to file containing 'spect_files' list of tuples
        saved by function make_spects_from_list_of_files.
        Default is None, in which case this function looks for
        a file named 'spect_files' in output_dir.
    output_dir : str
        path to output_dir containing .spect files. Default is '.' (current
        working directory).
    dict_name : str
        prefix to filename '_data_dict' in which the dictionary is saved.
        Default is 'test'.

    Returns
    -------
    data_dict_path

    Function saves 'data_dict' file in output_dir, with following structure:
        spects : list
            of ndarray, spectrograms from audio files
        filenames : list
            same length as spects, filename of each audio file that was converted to spectrogram
        freq_bins : ndarray
            vector of frequencies where each value is a bin center. Same for all spectrograms
        time_bins : list
            of ndarrays, each a vector of times where each value is a bin center.
            One for each spectrogram
        labelset : list
            of strings, labels corresponding to each spectrogram
        labeled_timebins : list
            of ndarrays, each same length as time_bins but value is a label for that bin.
            In other words, the labels vector is mapped onto the time_bins vector for the
            spectrogram.
        X : ndarray
            X_train, X_val, or X_test, depending on which data_dict you are looking at.
            Some number of spectrograms concatenated, enough so that the total duration
            of the spectrogram in time bins is equal to or greater than the target duration.
            If greater than target, then X is truncated so it is equal to the target.
        Y : ndarray
            Concatenated labeled_timebins vectors corresponding to the spectrograms in X.
        spect_ID_vector : ndarray
            Vector where each element is an ID for a song. Used to randomly grab subsets
            of data of a target duration while still having the subset be composed of
            individual songs as much as possible. So this vector will look like:
            [0, 0, 0, ..., 1, 1, 1, ... , n, n, n] where n is equal to or (a little) less
            than the length of spects. spect_ID_vector.shape[-1] is the same as X.shape[-1]
            and Y.shape[0].
        timebin_dur : float
            duration of a timebin in seconds from spectrograms
        spect_params : dict
            parameters for computing spectrogram as specified in config.ini file.
            Will be checked against .ini file when running other cli such as learn_curve.py
        labels_mapping : dict
            maps str labels for syllables to consecutive integers.
            As explained in docstring for make_spects_from_list_of_files.
    """
    spect_list = joblib.load(spect_files)
    spect_files = [tup[0] for tup in spect_list]

    labels = itertools.chain.from_iterable(
        [tup[2] for tup in spect_list])
    labels = set(labels)

    if labels != set(labelset):
        raise ValueError(
            'labels from all spect_files are not consistent with '
            'labels in labelset.')

    spects = []
    filenames = []
    all_time_bins = []
    labels = []
    labeled_timebins = []
    spect_ID_vector = []

    for spect_ind, tup in enumerate(spect_list):
        spect_dict = joblib.load(tup[0])
        spects.append(spect_dict['spect'])
        filenames.append(tup[0])
        all_time_bins.append(spect_dict['time_bins'])
        labels.append(spect_dict['labels'])
        labeled_timebins.append(spect_dict['labeled_timebins'])
        spect_ID_vector.extend([spect_ind] * spect_dict['time_bins'].shape[-1])

        if 'freq_bins' in locals():
            assert np.array_equal(spect_dict['freq_bins'], freq_bins)
        else:
            freq_bins = spect_dict['freq_bins']

        if 'labels_mapping' in locals():
            assert spect_dict['labels_mapping'] == labels_mapping
        else:
            labels_mapping = spect_dict['labels_mapping']

        if 'timebin_dur' in locals():
            assert spect_dict['timebin_dur'] == timebin_dur
        else:
            timebin_dur = spect_dict['timebin_dur']

        if 'spect_params' in locals():
            assert spect_dict['spect_params'] == spect_params
        else:
            spect_params = spect_dict['spect_params']

    X = np.concatenate(spects, axis=1)
    Y = np.concatenate(labeled_timebins)
    spect_ID_vector = np.asarray(spect_ID_vector, dtype='int')
    assert X.shape[-1] == Y.shape[0]  # Y has shape (timebins, 1)

    data_dict = {'spects': spects,
                 'filenames': filenames,
                 'freq_bins': freq_bins,
                 'time_bins': all_time_bins,
                 'labels': labels,
                 'labeled_timebins': labeled_timebins,
                 'spect_ID_vector': spect_ID_vector,
                 'X_' + dict_name: X,
                 'Y_' + dict_name: Y,
                 'timebin_dur': timebin_dur,
                 'spect_params': spect_params,
                 'labels_mapping': labels_mapping}

    data_dict_path = os.path.join(output_dir, dict_name + '_data_dict')
    print('saving data dictionary {} in {}'
          .format(data_dict_path, os.path.abspath(output_dir)))
    joblib.dump(data_dict, data_dict_path)
    return data_dict_path


def get_inds_for_dur(spect_ID_vector,
                     labeled_timebins_vector,
                     labels_mapping,
                     target_duration,
                     timebin_dur_in_s=0.001,
                     max_iter=1000,
                     method='incfreq'):
    """Randomly draw a subset of data with a specific duration.
    Draws spectrograms at random and adds to list until total duration of
    all spectrograms => target_duration, and then truncates at target duration.
    Returns vector of indices to select randomly-drawn subset from total dataset.

    Parameters
    ----------
    spect_ID_vector : ndarray
        vector where each element is an ID for a song. Used to randomly grab subsets
        of data of a target duration while still having the subset be composed of
        individual spectrograms as much as possible. So this vector will look like:
        [0, 0, 0, ..., 1, 1, 1, ... , n, n, n] where each n is the ID for one
        spectrogram, i.e. one audio file.
        spect_ID_vector.shape[-1] is the same as spectrograms.shape[-1] and labels.shape[0].
    labeled_timebins_vector : numpy.ndarray
        vector of same length as spect_ID_vector but each element is a label for the timebin,
        i.e. one of the possible classes.
    labels_mapping : dict
        maps string labels to consecutive integers. Used to check that the randomly drawn
        data set contains all classes.
    target_duration : float
        target duration of training set in seconds.
    timebin_dur_in_s : float
        duration of each timebin, i.e. each column in spectrogram, in seconds.
        Default is 0.001 s (1 ms)
    max_iter : int
        number of iterations to try drawing random subset of song that contains
        all classes in labels mapping.
        Default is 1000.
    method : str
        {'rand', 'incfreq'}
        method by which to obtain subset from training set.
        'incfreq' grabs songs at random but starts from the subset
        that includes the least frequently occurring class. Continues
        to grab randomly in order of increasing frequency until all
        classes are present, and then goes back to 'rand' method.
        'rand' grabs songs totally at random.
        Default is 'incfreq'.

    Returns
    -------
    inds_to_use : numpy.ndarray
        numpy array of indices, to index into rows of X_train that belong to chosen subset
        (assumes X_train is one long spectrogram, consisting of all
        training spectrograms concatenated, and each row being one timebin)
    """
    labeled_timebins_vector = np.squeeze(labeled_timebins_vector)

    if labeled_timebins_vector.ndim > 1:
        raise ValueError('number of dimensions of labeled_timebins_vector should be 1 '
                         '(after np.squeeze), but was equal to {}'
                         .format(labeled_timebins_vector.ndim))

    iter = 1
    with tqdm(total=max_iter) as pbar:
        while 1:  # keep iterating until we randomly draw subset that meets our criteria
            pbar.set_description(
                f'Randomly drawing subset of training data, attempt {iter} of {max_iter}'
            )
            if 'inds_to_use' in locals():
                del inds_to_use

            spect_IDs, spect_timebins = np.unique(spect_ID_vector, return_counts=True)

            if iter == 1:
                # sanity check:
                # spect_IDs should always start from 0
                # and go to n-1 where n is # of spectrograms
                assert np.array_equal(spect_IDs, np.arange(spect_IDs.shape[-1]))

            spect_IDs = spect_IDs.tolist()  # because we need to pop off ids for 'incfreq'

            spect_IDs_in_subset = []
            total_dur_in_timebins = 0

            if method == 'incfreq':
                classes, counts = np.unique(labeled_timebins_vector, return_counts=True)
                int_labels_without_int_flag = [val for val in labels_mapping.values()
                                               if type(val) is int]
                if set(classes) != set(int_labels_without_int_flag):
                    raise ValueError('classes in labeled_timebins_vector '
                                     'do not match classes in labels_mapping.')
                freq_rank = np.argsort(counts).tolist()

                # reason for doing it in this Schliemel-the-painter-looking way is that
                # I want to make sure all classes are represented first, but then
                # go back to just grabbing songs completely at random
                while freq_rank:  # is not an empty list yet
                    curr_class = classes[freq_rank.pop(0)]

                    # if curr_class already represented in subset, skip it
                    if 'inds_to_use' in locals():
                        classes_already_in_subset = np.unique(
                            labeled_timebins_vector[inds_to_use])
                        if curr_class in classes_already_in_subset:
                            continue

                    inds_this_class = np.where(labeled_timebins_vector==curr_class)[0]
                    spect_IDs_this_class = np.unique(spect_ID_vector[inds_this_class])
                    # keep only the spect IDs we haven't popped off main list already
                    spect_IDs_this_class = [spect_ID_this_class
                                            for spect_ID_this_class in spect_IDs_this_class
                                            if spect_ID_this_class in spect_IDs]
                    rand_spect_ID = np.random.choice(spect_IDs_this_class)

                    spect_IDs_in_subset.append(rand_spect_ID)
                    spect_IDs.pop(spect_IDs.index(rand_spect_ID))  # so as not to reuse it
                    # below, [0] because np.where returns tuple
                    spect_ID_inds = np.where(spect_ID_vector == rand_spect_ID)[0]
                    if 'inds_to_use' in locals():
                        inds_to_use = np.concatenate((inds_to_use, spect_ID_inds))
                    else:
                        inds_to_use = spect_ID_inds
                    total_dur_in_timebins += spect_timebins[rand_spect_ID]
                    if total_dur_in_timebins * timebin_dur_in_s >= target_duration:
                        # if total_dur greater than target, need to truncate
                        if total_dur_in_timebins * timebin_dur_in_s > target_duration:
                            correct_length = np.round(target_duration / timebin_dur_in_s).astype(int)
                            inds_to_use = inds_to_use[:correct_length]
                        # (if equal to target, don't need to do anything)
                        break

            if method=='rand' or \
                    total_dur_in_timebins * timebin_dur_in_s < target_duration:

                shuffled_spect_IDs = np.random.permutation(spect_IDs)

                for spect_ID in shuffled_spect_IDs:
                    spect_IDs_in_subset.append(spect_ID)
                    # below, [0] because np.where returns tuple
                    spect_ID_inds = np.where(spect_ID_vector == spect_ID)[0]
                    if 'inds_to_use' in locals():
                        inds_to_use = np.concatenate((inds_to_use, spect_ID_inds))
                    else:
                        inds_to_use = spect_ID_inds
                    total_dur_in_timebins += spect_timebins[spect_ID]
                    if total_dur_in_timebins * timebin_dur_in_s >= target_duration:
                        # if total_dur greater than target, need to truncate
                        if total_dur_in_timebins * timebin_dur_in_s > target_duration:
                            correct_length = np.round(target_duration / timebin_dur_in_s).astype(int)
                            inds_to_use = inds_to_use[:correct_length]
                        # (if equal to target, don't need to do anything)
                        break

            if set(np.unique(labeled_timebins_vector[inds_to_use])) != set(labels_mapping.values()):
                pbar.update(1)
                iter += 1

                if iter > max_iter:
                    raise ValueError('Attempted to draw subset of songs at random '
                                     'that contained all classes '
                                     'more than {} times unsuccessfully. Make sure '
                                     'that all classes are present in training data '
                                     'set.'.format(max_iter))
            else:
                return inds_to_use


def reshape_data_for_batching(X, batch_size, time_steps, Y=None):
    """Reshape data to feed to network in batches.
    Data is returned with shape (batch_size, num_batches * time_steps, -1)
    so that a batch can be grabbed at random starting at any index on axis 1
    (that is less than "length of axis 1 minus time steps").
    Pads data with zeros if it cannot be evenly divided into batches of size batch_size.

    Parameters
    ----------
    X : numpy.ndarray
        2-d matrix, concatenated spectrograms.
        Spectrograms are oriented so rows are time bins and columns are frequency bins.
    batch_size : int
        Number of samples in a batch
    time_steps : int
        Number of time steps in each sample, i.e., number of rows
    Y : numpy.ndarray
        vector of labels with shape (X.shape[0], 1). Default is None, for the case when
        you only have spectrograms in X and need to reshape them
        so you can predict labels Y.

    Returns
    -------
    X : numpy.ndarray
        Spectrograms reshaped to have shape (batch_size, num_batches * time_steps, -1).
    Y : numpy.ndarray
        Labels reshaped, will have shape (batch_size, -1). Only returned if Y is provided as an argument.
    num_batches : int
        num_batches = X.shape[0] // batch_size // time_steps (plus 1, if zero-padding was required).
    """
    num_batches = X.shape[0] // batch_size // time_steps
    rows_to_append = ((num_batches + 1) * time_steps * batch_size) - X.shape[0]

    if rows_to_append > 0:
        X = np.concatenate((X, np.zeros((rows_to_append, X.shape[1]))),
                           axis=0)
        if Y is not None:
            Y = np.concatenate((Y, np.zeros((rows_to_append, 1), dtype=int)), axis=0)
        num_batches = num_batches + 1

    X = X.reshape((batch_size, num_batches * time_steps, -1))
    if Y is not None:
        Y = Y.reshape((batch_size, -1))

    if Y is not None:
        return X, Y, num_batches
    else:
        return X, num_batches


def convert_timebins_to_labels(labeled_timebins,
                               labels_mapping,
                               spect_ID_vector=None):
    """converts output of cnn-bilstm from label for each frame
    to one label for each continuous segment

    Parameters
    ----------
    labeled_timebins : ndarray
        where each element is a label for a time bin.
        Such an array is the output of the cnn-bilstm network.
    labels_mapping : dict
        that maps str labels to consecutive integers.
        The mapping is inverted to convert back to str labels.
    spect_ID_vector : ndarray
        of same length as labeled_timebins, where each element
        is an ID # for the spectrogram from which labeled_timebins
        was taken.
        If provided, used to split the converted labels back to
        a list of label str, with one for each spectrogram.
        Default is None, in which case the return value is one long str.

    Returns
    -------
    labels : str or list
        labeled_timebins mapped back to label str.
        If spect_ID_vector was provided, then labels is split into a list of str,
        where each str corresponds to predicted labels for each predicted
        segment in each spectrogram as identified by spect_ID_vector.
    """

    idx = np.diff(labeled_timebins, axis=0).astype(np.bool)
    idx = np.insert(idx, 0, True)

    labels = labeled_timebins[idx]

    # remove silent gap label
    silent_gap_label = labels_mapping['silent_gap_label']
    labels = labels[labels != silent_gap_label]
    labels = labels.tolist()

    inverse_labels_mapping = dict((v, k) for k, v
                                  in labels_mapping.items())
    labels = [inverse_labels_mapping[label] for label in labels]

    if spect_ID_vector:
        labels_list = []
        spect_ID_vector = spect_ID_vector[idx]
        labels_arr = np.asarray(labels)
        # need to split up labels by spect_ID_vector
        # this is probably not the most efficient way:
        spect_IDs = np.unique(spect_ID_vector)

        for spect_ID in spect_IDs:
            these = np.where(spect_ID_vector == spect_ID)
            curr_labels = labels_arr[these].tolist()
            if all([type(el) is str for el in curr_labels]):
                labels_list.append(''.join(curr_labels))
            elif all([type(el) is int for el in curr_labels]):
                labels_list.append(curr_labels)
        return labels_list, spect_ID_vector
    else:
        if all([type(el) is str for el in labels]):
            return ''.join(labels)
        elif all([type(el) is int for el in labels]):
            return labels


def range_str(range_str, sort=True):
    """Generate range of ints from a formatted string,
    then convert range from int to str

    Example:
        >>> range_str('1-4,6,9-11')
        ['1','2','3','4','6','9','10','11']

    Takes a range in form of "a-b" and returns
    a list of numbers between a and b inclusive.
    Also accepts comma separated ranges like "a-b,c-d,f"  which will
    return a list with numbers from a to b, c to d, and f.

    Parameters
    ----------
    range_str : str
        of form 'a-b,c'
        where a hyphen indicates a range
        and a comma separates ranges or single numbers
    sort : bool
        If True, sort output before returning. Default is True.

    Returns
    -------
    list_range : list
        of int, produced by parsing range_str
    """
    # adapted from
    # http://code.activestate.com/recipes/577279-generate-list-of-numbers-from-hyphenated-and-comma/
    s = "".join(range_str.split())  # removes white space
    list_range = []
    for substr in range_str.split(','):
        subrange = substr.split('-')
        if len(subrange) not in [1, 2]:
            raise SyntaxError("unable to parse range {} in labelset {}."
                              .format(subrange, substr))
        list_range.extend(
            [int(subrange[0])]
        ) if len(subrange) == 1 else list_range.extend(
            range(int(subrange[0]), int(subrange[1]) + 1))

    if sort:
        list_range.sort()

    return [str(list_int) for list_int in list_range]


from . import config
from . import __main__
from . import curvefit
from . import evfuncs
from .utils import mat
from . import metrics
from . import utils
from . import koumura_utils
from . import cli
from . import network



"""parses [OUTPUT] section of config"""
import os
from configparser import NoOptionError

import attr
from attr.validators import optional

from .validators import is_a_directory


@attr.s
class OutputConfig:
    """class that represents [OUTPUT] section of config.ini file

    Attributes
    ----------
    root_results_dir : str
        directory in which results *will* be created.
        The vak.cli.learcurve function will create
        a subdirectory in this directory each time it runs.
    results_dirname : str
        name of subdirectory created by vak.cli.learncurve.
        This option is added programatically by that function
        when it runs but can be changed (e.g. to run
        vak.cli.summary on previous outputs of learncurve.)
    """
    root_results_dir = attr.ib(validator=optional(is_a_directory), default=None)
    results_dirname = attr.ib(validator=optional(is_a_directory), default=None)


def parse_output_config(config):
    """parse [OUTPUT] section of config.ini file

    Parameters
    ----------
    config : ConfigParser
        containing config.ini file already loaded by parse function

    Returns
    -------
    output_config : namedtuple
    """
    try:
        root_results_dir = config['OUTPUT']['root_results_dir']
        root_results_dir = os.path.expanduser(root_results_dir)
    except NoOptionError:
        raise KeyError('must specify root_results_dir in [OUTPUT] section '
                       'of config.ini file')

    if config.has_option('OUTPUT', 'results_dir_made_by_main_script'):
        # don't check whether it exists because it may not yet,
        # depending on which function we are calling.
        # So it's up to calling function to check for existence of directory
        results_dirname = config['OUTPUT']['results_dir_made_by_main_script']
        results_dirname = os.path.expanduser(results_dirname)
    else:
        results_dirname = None

    return OutputConfig(root_results_dir,
                        results_dirname)



"""parses [DATA] section of config"""
import os

import attr
from attr.validators import instance_of, optional

from ..utils.data import range_str
from .validators import is_a_directory, is_a_file


@attr.s
class DataConfig:
    """class to represent [DATA] section of config.ini file

    Attributes
    ----------
    labelset : list
        of str or int, set of labels for syllables
    all_labels_are_int : bool
        if True, labels are of type int, not str
    silent_gap_label  : int
        label for time bins of silent gaps between syllables.
        Type is int because labels are converted to a set of
        n consecutive integers {0,1,2...n} where n is the number
        of syllable classes + the silent gap class.
        Default is 0 (in which case labels are {1,2,3,...,n}).
    skip_files_with_labels_not_in_labelset : bool
        if True, skip a file if the labels variable contains labels not
        found in 'labelset'. Default is True.
    output_dir : str
        Path to location where data sets should be saved. Default is None,
        in which case data sets are saved in the current working directory.
    mat_spect_files_path : str
        Path to a directory of .mat files that contain spectrograms.
        Default is None (and this implies user is supplying audio files
         instead of supplying spectrograms in .mat files).
    mat_spects_annotation_file : str
        Path to annotation file associated with .mat files.
        Default is None.
    data_dir : str
        path to directory with audio files from which to make dataset
    total_train_set_dur : float
        total duration of training set, in seconds.
        Training subsets of shorter duration will be drawn from this set.
    val_dur : float
        total duration of validation set, in seconds.
    test_dur : float
        total duration of test set, in seconds.
    save_transformed_data : bool
        if True, save transformed data (i.e. scaled, reshaped). The data can then
        be used on a subsequent run (e.g. if you want to compare results
        from different hyperparameters across the exact same training set).
        Also useful if you need to check what the data looks like when fed to networks.
        Default is False.
    """
    labelset = attr.ib(validator=instance_of(list))

    total_train_set_dur = attr.ib(validator=optional(instance_of(float)), default=None)
    val_dur = attr.ib(validator=optional(instance_of(float)), default=None)
    test_dur = attr.ib(validator=optional(instance_of(float)), default=None)
    all_labels_are_int = attr.ib(validator=instance_of(bool), default=False)
    silent_gap_label = attr.ib(validator=instance_of(int), default=0)
    skip_files_with_labels_not_in_labelset = attr.ib(validator=instance_of(bool), default=True)
    output_dir = attr.ib(validator=optional(is_a_directory), default=None)
    mat_spect_files_path = attr.ib(validator=optional(is_a_directory), default=None)
    mat_spects_annotation_file = attr.ib(validator=optional(is_a_file), default=None)
    data_dir = attr.ib(validator=optional(is_a_directory), default=None)
    save_transformed_data = attr.ib(validator=instance_of(bool), default=False)


def parse_data_config(config, config_file):
    """parse [DATA] section of config.ini file

    Parameters
    ----------
    config : ConfigParser
        containing config.ini file already loaded by parse function
    config_file : str
        path to config file (used for error messages)

    Returns
    -------
    data_config : vak.config.data.DataConfig
        instance of class that represents [DATA] section of config.ini file
    """
    config_dict = {}

    labelset = config['DATA']['labelset']
    # make mapping from syllable labels to consecutive integers
    # start at 1, because 0 is assumed to be label for silent gaps
    if '-' in labelset or ',' in labelset:
        # if user specified range of ints using a str
        config_dict['labelset'] = range_str(labelset)
    else:  # assume labelset is characters
        config_dict['labelset'] = list(labelset)

    # to make type-checking consistent across .mat / .cbin / Koumura .wav files
    # set all_labels_are_int flag
    # currently only used with .mat files
    if config.has_option('DATA', 'all_labels_are_int'):
        config_dict['all_labels_are_int'] = config.getboolean('DATA', 'all_labels_are_int')

    if config.has_option('DATA', 'silent_gap_label'):
        config_dict['silent_gap_label'] = int(config['DATA']['silent_gap_label'])

    if config.has_option('DATA', 'skip_files_with_labels_not_in_labelset'):
        config_dict[
            'skip_files_with_labels_not_in_labelset'
        ] = config.getboolean('DATA',
                              'skip_files_with_labels_not_in_labelset')

    if config.has_option('DATA', 'output_dir'):
        output_dir = config['DATA']['output_dir']
        output_dir = os.path.expanduser(output_dir)
        config_dict['output_dir'] = os.path.abspath(output_dir)

    # if using spectrograms from .mat files
    if config.has_option('DATA', 'mat_spect_files_path'):
        # make spect_files file from .mat spect files and annotation file
        config_dict['mat_spect_files_path'] = config['DATA']['mat_spect_files_path']
        config_dict['mat_spects_annotation_file'] = config['DATA']['mat_spect_files_annotation_file']

    if config.has_option('DATA', 'data_dir'):
        data_dir = config['DATA']['data_dir']
        config_dict['data_dir'] = os.path.expanduser(data_dir)

    if config.has_option('DATA', 'total_train_set_duration'):
        config_dict['total_train_set_dur'] = float(config['DATA']['total_train_set_duration'])

    if config.has_option('DATA', 'validation_set_duration'):
        config_dict['val_dur'] = float(config['DATA']['validation_set_duration'])

    if config.has_option('DATA', 'test_set_duration'):
        config_dict['test_dur'] = float(config['DATA']['test_set_duration'])

    if config.has_option('DATA', 'save_transformed_data'):
        config_dict['save_transformed_data'] = config.getboolean('DATA', 'save_transformed_data')

    return DataConfig(**config_dict)



"""
Invokes __main__ when the module is run as a script.
Example: python -m vak --help
"""
import argparse
import os
from glob import glob

from .cli import cli


def get_parser():
    """returns ArgumentParser instance used by main()"""
    CHOICES = [
        'prep',
        'train',
        'predict',
        'finetune',
        'learncurve',
        'summary',
    ]

    parser = argparse.ArgumentParser(description='vak command-line interface',
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('command', type=str, metavar='command',
                        choices=CHOICES,
                        help="Command to run, valid options are:\n"
                             f"{CHOICES}\n"
                             "$ vak train ./configs/config_2018-12-17.ini")
    parser.add_argument('configfile', type=str,
                        help='name of config.ini file to use \n'
                             '$ vak train ./configs/config_2018-12-17.ini')
    parser.add_argument('-d', '--dataset', type=str,
                        help='Create a dataset from a list of files '
                             'in a .txt file, by passing in the name of the .txt file.\n'
                             '$ vak --dataset ./audio_files_list.txt')
    parser.add_argument('-g', '--glob', type=str,
                        help='string to use with glob function '
                             'to search for config files fitting some pattern.\n'
                             '$ vak --glob ./config_finches*.ini')
    parser.add_argument('-t', '--txt', type=str,
                        help='name of .txt file containing list of '
                             'config files to run\n'
                             '$ vak --text ./list_of_config_filenames.txt')
    return parser


def main():
    """main function, called when package is run with `python -m vak` or
    alternatively just calling `vak` at the command line (because this
    function is installed under just `vak` as a console script entry point,
    see setup.py)
    """
    parser = get_parser()
    args = parser.parse_args()
    # logic for handling glob + txt flags, affects how conifg

    if args.glob:
        config_files = glob(args.glob)
    elif args.txt:
        with open(args.txt, 'r') as config_list_file:
            config_files = config_list_file.readlines()
    elif args.configfile:
        config_files = [args.configfile]  # single-item list

    if args.glob or args.txt:
        config_files_cleaned = []
        for config_file in config_files:
            config_file = config_file.rstrip()
            config_file = config_file.lstrip()
            if os.path.isfile(config_file):
                config_files_cleaned.append(config_file)
            else:
                if args.txt:
                    txt_dir = os.path.dirname(args.txt)
                    txt_dir_config = os.path.join(txt_dir,
                                                  config_file)
                    if os.path.isfile(txt_dir_config):
                        config_files_cleaned.append(txt_dir_config)
                    else:
                        raise FileNotFoundError("Can't find config file: {}"
                                                .format(txt_dir_config))
                else:  # if --glob was used instead of --txt
                    raise FileNotFoundError("Can't find config file: {}"
                                            .format(config_file))
        config_files = config_files_cleaned

    cli(command=args.command,
        config_files=config_files)


if __name__ == "__main__":
    main()



"""command-line interface functions for training,
creating learning curves, etc."""

from .prep import make_data
from .learncurve import learncurve
from .summary import summary
from .predict import predict
from .cli import cli
from .train import train



import os
import pickle
import sys
from datetime import datetime
from glob import glob

import joblib
import numpy as np
import tensorflow as tf

from .. import metrics, utils
import vak.network


def summary(results_dirname,
            train_data_dict_path,
            networks,
            train_set_durs,
            num_replicates,
            labelset,
            test_data_dict_path,
            normalize_spectrograms=False):
    """generate summary learning curve from networks trained by cli.learncurve
    Computes error on test set for each network trained by learncurve,
    and saves in a ./summary directory within results_dir

    Parameters
    ----------
    results_dirname : str
        path to directory containing results created by a run of learncurve
    train_data_dict_path : str
        path to training data
    networks : dict
        where each key is the name of a neural network and the corresponding
        value is the configuration for that network (in a namedtuple or a dict)
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20]
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate mean accuracy for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    labelset : list
        of str or int, set of labels for syllables
    test_data_dict_path : str
        path to test data
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.

    Returns
    -------
    None
    """
    if not os.path.isdir(results_dirname):
        raise FileNotFoundError('directory {}, specified as '
                                'results_dir_made_by_main_script, is not found.'
                                .format(results_dirname))
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    summary_dirname = os.path.join(results_dirname,
                                   'summary_' + timenow)
    os.makedirs(summary_dirname)

    labels_mapping_file = os.path.join(results_dirname, 'labels_mapping')
    with open(labels_mapping_file, 'rb') as labels_map_file_obj:
        labels_mapping = pickle.load(labels_map_file_obj)

    train_data_dict = joblib.load(train_data_dict_path)
    (X_train,
     Y_train,
     train_timebin_dur,
     train_spect_params,
     train_labels) = (train_data_dict['X_train'],
                      train_data_dict['Y_train'],
                      train_data_dict['timebin_dur'],
                      train_data_dict['spect_params'],
                      train_data_dict['labels'])
    labels_mapping = train_data_dict['labels_mapping']
    n_syllables = len(labels_mapping)
    X_train = X_train.T

    # only get this just to have in summary file if needed
    if all(type(labels_el) is str for labels_el in train_labels):
        # when taken from .not.mat files associated with .cbin audio files
        Y_train_labels = ''.join(train_labels)
        Y_train_labels_for_lev = Y_train_labels
    elif all(type(labels_el) is np.ndarray for labels_el in train_labels):
        # when taken from annotation.mat supplied with .wav audio files
        Y_train_labels = np.concatenate(train_labels).tolist()
        Y_train_labels_for_lev = ''.join([chr(lbl) for lbl in Y_train_labels])
    elif all(type(labels_el) is list for labels_el in train_labels):
        # when taken from annotation.xml supplied with Koumura .wav audio files
        Y_train_labels = [lbl for lbl_list in train_labels for lbl in lbl_list]
        if all([type(lbl) is int for lbl in Y_train_labels]):
            Y_train_labels_for_lev = ''.join(
                [chr(lbl) for lbl in Y_train_labels])
        elif all([type(lbl) is str for lbl in Y_train_labels]):
            Y_train_labels_for_lev = ''.join(Y_train_labels)
        else:
            raise TypeError('Couldn\'t determine type for training labels in {}'
                            .format(type(train_data_dict_path)))
    else:
        raise TypeError('Not able to determine type of labels in train data')

    print('loading testing data')
    test_data_dict = joblib.load(test_data_dict_path)

    # notice data is called `X_test_copy` and `Y_test_copy`
    # because main loop below needs a copy of the original
    # to normalize and reshape
    (X_test_copy,
     Y_test_copy,
     test_timebin_dur,
     files_used,
     test_spect_params,
     test_labels) = (test_data_dict['X_test'],
                     test_data_dict['Y_test'],
                     test_data_dict['timebin_dur'],
                     test_data_dict['filenames'],
                     test_data_dict['spect_params'],
                     test_data_dict['labels'])

    if train_spect_params != test_spect_params:
        raise ValueError('Spectrogram parameters for training data do not match those '
                         'for test data, will give incorrect error rate.')
    if train_timebin_dur != test_timebin_dur:
        raise ValueError('Durations of time bins in spectrograms for training data '
                         'does not match that of test data, will give incorrect error rate.')

    # have to transpose X_test so rows are timebins and columns are frequencies
    X_test_copy = X_test_copy.T
    if X_train.shape[-1] != X_test_copy.shape[-1]:
        raise ValueError(f'Number of frequency bins in training set spectrograms, {X_train.shape[-1]}, '
                         f'does not equal number in test set spectrograms, {X_test_copy.shape[-1]}.')
    freq_bins = X_test_copy.shape[-1]  # number of columns

    # save test set so it's clear from results directory alone
    # which test set was used
    joblib.dump(X_test_copy, os.path.join(results_dirname, 'X_test'))
    joblib.dump(Y_test_copy, os.path.join(results_dirname, 'Y_test'))

    # used for Levenshtein distance + syllable error rate
    if all(type(labels_el) is str for labels_el in test_labels):
        # when taken from .not.mat files associated with .cbin audio files
        Y_test_labels = ''.join(test_labels)
        Y_test_labels_for_lev = Y_test_labels
    elif all(type(labels_el) is np.ndarray for labels_el in test_labels):
        # when taken from annotation.mat supplied with .wav audio files
        Y_test_labels = np.concatenate(test_labels).tolist()
        Y_test_labels_for_lev = ''.join([chr(lbl) for lbl in Y_test_labels])
    elif all(type(labels_el) is list for labels_el in test_labels):
        # when taken from annotation.xml supplied with Koumura .wav audio files
        Y_test_labels = [lbl for lbl_list in test_labels for lbl in lbl_list]
        if all([type(lbl) is int for lbl in Y_train_labels]):
            Y_test_labels_for_lev = ''.join(
                [chr(lbl) for lbl in Y_test_labels])
        elif all([type(lbl) is str for lbl in Y_train_labels]):
            Y_test_labels_for_lev = ''.join(Y_test_labels)
        else:
            raise TypeError('Couldn\'t determine type for test labels in {}'
                            .format(type(test_data_dict_path)))
    else:
        raise TypeError('Not able to determine type of labels in test data')

    replicates = range(num_replicates)
    # initialize arrays to hold summary results
    Y_pred_test_all = []  # will be a nested list
    Y_pred_train_all = []  # will be a nested list
    Y_pred_test_labels_all = []
    Y_pred_train_labels_all = []
    train_err_arr = np.empty((len(train_set_durs), len(replicates)))
    test_err_arr = np.empty((len(train_set_durs), len(replicates)))
    train_lev_arr = np.empty((len(train_set_durs), len(replicates)))
    test_lev_arr = np.empty((len(train_set_durs), len(replicates)))
    train_syl_err_arr = np.empty((len(train_set_durs), len(replicates)))
    test_syl_err_arr = np.empty((len(train_set_durs), len(replicates)))

    NETWORKS = vak.network._load()

    for dur_ind, train_set_dur in enumerate(train_set_durs):

        Y_pred_test_this_dur = []
        Y_pred_train_this_dur = []
        Y_pred_test_labels_this_dur = []
        Y_pred_train_labels_this_dur = []

        for rep_ind, replicate in enumerate(replicates):
            print("getting train and test error for "
                  "training set with duration of {} seconds, "
                  "replicate {}".format(train_set_dur, replicate))

            training_records_dir = ('records_for_training_set_with_duration_of_'
                                    + str(train_set_dur) + '_sec_replicate_'
                                    + str(replicate))
            training_records_path = os.path.join(results_dirname,
                                                 training_records_dir)

            train_inds_file = glob(os.path.join(training_records_path, 'train_inds'))
            if len(train_inds_file) != 1:
                raise ValueError("incorrect number of train_inds files in {}, "
                                 "should only be one but found: {}"
                                 .format(training_records_path, train_inds_file))
            else:
                train_inds_file = train_inds_file[0]
            with open(os.path.join(train_inds_file), 'rb') as train_inds_file:
                train_inds = pickle.load(train_inds_file)

            # get training set
            Y_train_subset = Y_train[train_inds]
            X_train_subset = X_train[train_inds, :]
            assert Y_train_subset.shape[0] == X_train_subset.shape[0], \
                "mismatch between X and Y train subset shapes"
            if normalize_spectrograms:
                scaler_name = ('spect_scaler_duration_{}_replicate_{}'
                               .format(train_set_dur, replicate))
                spect_scaler = joblib.load(
                    os.path.join(training_records_path, scaler_name))
                X_train_subset = spect_scaler.transform(X_train_subset)

            # Normalize before reshaping to avoid even more convoluted array reshaping.
            if normalize_spectrograms:
                X_test = spect_scaler.transform(X_test_copy)
            else:
                # get back "un-reshaped" X_test
                X_test = np.copy(X_test_copy)

            # need to get Y_test from copy because it gets reshaped every time through loop
            Y_test = np.copy(Y_test_copy)

            # save scaled spectrograms. Note we already saved training data scaled
            scaled_test_data_filename = os.path.join(summary_dirname,
                                                     'scaled_test_spects_duration_{}_replicate_{}'
                                                     .format(train_set_dur,
                                                             replicate))
            scaled_test_data_dict = {'X_test_scaled': X_test}
            joblib.dump(scaled_test_data_dict, scaled_test_data_filename)

            for net_name, net_config in networks.items():
                # reload network #
                net_config_dict = net_config._asdict()
                net_config_dict['n_syllables'] = n_syllables
                if 'freq_bins' in net_config_dict:
                    net_config_dict['freq_bins'] = freq_bins
                net = NETWORKS[net_name](**net_config_dict)

                results_dirname_this_net = os.path.join(training_records_path, net_name)

                checkpoint_path = os.path.join(results_dirname_this_net, 'checkpoints')
                # we use latest checkpoint when doing summary for learncurve, assume that's "best trained"
                checkpoint_file = tf.train.latest_checkpoint(checkpoint_dir=checkpoint_path)

                meta_file = glob(checkpoint_file + '*meta')
                if len(meta_file) != 1:
                    raise ValueError('Incorrect number of meta files for last saved checkpoint.\n'
                                     'For checkpoint {}, found these files:\n'
                                     '{}'
                                     .format(checkpoint_file, meta_file))
                else:
                    meta_file = meta_file[0]

                data_file = glob(checkpoint_file + '*data*')
                if len(data_file) != 1:
                    raise ValueError('Incorrect number of data files for last saved checkpoint.\n'
                                     'For checkpoint {}, found these files:\n'
                                     '{}'
                                     .format(checkpoint_file, data_file))
                else:
                    data_file = data_file[0]

                # reshape data for batching using net_config #
                (X_train_subset,
                 Y_train_subset,
                 num_batches_train) = utils.data.reshape_data_for_batching(
                    X_train_subset,
                    net_config.batch_size,
                    net_config.time_bins,
                    Y_train_subset)

                (X_test,
                 Y_test,
                 num_batches_test) = utils.data.reshape_data_for_batching(
                    X_test,
                    net_config.batch_size,
                    net_config.time_bins,
                    Y_test)

                scaled_reshaped_data_filename = os.path.join(summary_dirname,
                                                             'scaled_reshaped_spects_duration_{}_replicate_{}'
                                                             .format(train_set_dur,
                                                                     replicate))
                scaled_reshaped_data_dict = {
                    'X_train_subset_scaled_reshaped': X_train_subset,
                    'Y_train_subset_reshaped': Y_train_subset,
                    'X_test_scaled_reshaped': X_test,
                    'Y_test_reshaped': Y_test}
                joblib.dump(scaled_reshaped_data_dict,
                            scaled_reshaped_data_filename)

                with tf.Session(graph=net.graph) as sess:
                    tf.logging.set_verbosity(tf.logging.ERROR)

                    net.restore(sess=sess,
                                meta_file=meta_file,
                                data_file=data_file)

                    if 'Y_pred_train' in locals():
                        del Y_pred_train

                    print('calculating training set error')
                    for b in range(num_batches_train):  # "b" is "batch number"
                        d = {net.X: X_train_subset[:,
                                    b * net_config.time_bins: (b + 1) * net_config.time_bins, :],
                             net.lng: [net_config.time_bins] * net_config.batch_size}

                        if 'Y_pred_train' in locals():
                            preds = sess.run(net.predict, feed_dict=d)
                            preds = preds.reshape(net_config.batch_size, -1)
                            Y_pred_train = np.concatenate((Y_pred_train, preds),
                                                          axis=1)
                        else:
                            Y_pred_train = sess.run(net.predict, feed_dict=d)
                            Y_pred_train = Y_pred_train.reshape(net_config.batch_size, -1)

                    Y_train_subset = Y_train[train_inds]  # get back "unreshaped" Y_train_subset
                    # get rid of predictions to zero padding that don't matter
                    Y_pred_train = Y_pred_train.ravel()[:Y_train_subset.shape[0],
                                   np.newaxis]
                    train_err = np.sum(Y_pred_train - Y_train_subset != 0) / \
                                Y_train_subset.shape[0]
                    train_err_arr[dur_ind, rep_ind] = train_err
                    print('train error was {}'.format(train_err))
                    Y_pred_train_this_dur.append(Y_pred_train)

                    Y_train_subset_labels = utils.data.convert_timebins_to_labels(Y_train_subset,
                                                                                  labels_mapping)
                    Y_pred_train_labels = utils.data.convert_timebins_to_labels(Y_pred_train,
                                                                                labels_mapping)
                    Y_pred_train_labels_this_dur.append(Y_pred_train_labels)

                    if all([type(el) is int for el in Y_train_subset_labels]):
                        # if labels are ints instead of str
                        # convert to str just to calculate Levenshtein distance
                        # and syllable error rate.
                        # Let them be weird characters (e.g. '\t') because that doesn't matter
                        # for calculating Levenshtein distance / syl err rate
                        Y_train_subset_labels = ''.join(
                            [chr(el) for el in Y_train_subset_labels])
                        Y_pred_train_labels = ''.join(
                            [chr(el) for el in Y_pred_train_labels])

                    train_lev = metrics.levenshtein(Y_pred_train_labels,
                                                    Y_train_subset_labels)
                    train_lev_arr[dur_ind, rep_ind] = train_lev
                    print('Levenshtein distance for train set was {}'.format(
                        train_lev))
                    train_syl_err_rate = metrics.syllable_error_rate(
                        Y_train_subset_labels,
                        Y_pred_train_labels)
                    train_syl_err_arr[dur_ind, rep_ind] = train_syl_err_rate
                    print('Syllable error rate for train set was {}'.format(
                        train_syl_err_rate))

                    if 'Y_pred_test' in locals():
                        del Y_pred_test

                    print('calculating test set error')
                    for b in range(num_batches_test):  # "b" is "batch number"
                        d = {
                            net.X: X_test[:, b * net_config.time_bins: (b + 1) * net_config.time_bins, :],
                            net.lng: [net_config.time_bins] * net_config.batch_size}

                        if 'Y_pred_test' in locals():
                            preds = sess.run(net.predict, feed_dict=d)
                            preds = preds.reshape(net_config.batch_size, -1)
                            Y_pred_test = np.concatenate((Y_pred_test, preds),
                                                         axis=1)
                        else:
                            Y_pred_test = sess.run(net.predict, feed_dict=d)
                            Y_pred_test = Y_pred_test.reshape(net_config.batch_size, -1)

                    # again get rid of zero padding predictions
                    Y_pred_test = Y_pred_test.ravel()[:Y_test_copy.shape[0],
                                  np.newaxis]
                    test_err = np.sum(Y_pred_test - Y_test_copy != 0) / \
                               Y_test_copy.shape[0]
                    test_err_arr[dur_ind, rep_ind] = test_err
                    print('test error was {}'.format(test_err))
                    Y_pred_test_this_dur.append(Y_pred_test)

                    Y_pred_test_labels = utils.data.convert_timebins_to_labels(Y_pred_test,
                                                                               labels_mapping)
                    Y_pred_test_labels_this_dur.append(Y_pred_test_labels)
                    if all([type(el) is int for el in Y_pred_test_labels]):
                        # if labels are ints instead of str
                        # convert to str just to calculate Levenshtein distance
                        # and syllable error rate.
                        # Let them be weird characters (e.g. '\t') because that doesn't matter
                        # for calculating Levenshtein distance / syl err rate
                        Y_pred_test_labels = ''.join(
                            [chr(el) for el in Y_pred_test_labels])
                        # already converted actual Y_test_labels from int to str above,
                        # stored in variable `Y_test_labels_for_lev`

                    test_lev = metrics.levenshtein(Y_pred_test_labels,
                                                   Y_test_labels_for_lev)
                    test_lev_arr[dur_ind, rep_ind] = test_lev
                    print(
                        'Levenshtein distance for test set was {}'.format(test_lev))
                    test_syl_err_rate = metrics.syllable_error_rate(
                        Y_test_labels_for_lev,
                        Y_pred_test_labels)
                    print('Syllable error rate for test set was {}'.format(
                        test_syl_err_rate))
                    test_syl_err_arr[dur_ind, rep_ind] = test_syl_err_rate

        Y_pred_train_all.append(Y_pred_train_this_dur)
        Y_pred_test_all.append(Y_pred_test_this_dur)
        Y_pred_train_labels_all.append(Y_pred_train_labels_this_dur)
        Y_pred_test_labels_all.append(Y_pred_test_labels_this_dur)

    Y_pred_train_filename = os.path.join(summary_dirname,
                                         'Y_pred_train_all')
    with open(Y_pred_train_filename, 'wb') as Y_pred_train_file:
        pickle.dump(Y_pred_train_all, Y_pred_train_file)

    Y_pred_test_filename = os.path.join(summary_dirname,
                                        'Y_pred_test_all')
    with open(Y_pred_test_filename, 'wb') as Y_pred_test_file:
        pickle.dump(Y_pred_test_all, Y_pred_test_file)

    train_err_filename = os.path.join(summary_dirname,
                                      'train_err')
    with open(train_err_filename, 'wb') as train_err_file:
        pickle.dump(train_err_arr, train_err_file)

    test_err_filename = os.path.join(summary_dirname,
                                     'test_err')
    with open(test_err_filename, 'wb') as test_err_file:
        pickle.dump(test_err_arr, test_err_file)

    pred_and_err_dict = {'Y_pred_train_all': Y_pred_train_all,
                         'Y_pred_test_all': Y_pred_test_all,
                         'Y_pred_train_labels_all': Y_pred_train_labels_all,
                         'Y_pred_test_labels_all': Y_pred_test_labels_all,
                         'Y_train_labels': Y_train_labels,
                         'Y_test_labels': Y_test_labels,
                         'train_err': train_err_arr,
                         'test_err': test_err_arr,
                         'train_lev': train_lev_arr,
                         'train_syl_err_rate': train_syl_err_arr,
                         'test_lev': test_lev_arr,
                         'test_syl_err_rate': test_syl_err_arr,
                         'train_set_durs': train_set_durs,
                         'num_replicates': num_replicates}

    pred_err_dict_filename = os.path.join(summary_dirname,
                                          'y_preds_and_err_for_train_and_test')
    joblib.dump(pred_and_err_dict, pred_err_dict_filename)


if __name__ == '__main__':
    config_file = sys.argv[1]
    summary(config_file)



import logging
import os
import sys
from configparser import ConfigParser
from datetime import datetime
from glob import glob

from ..utils import make_spects_from_list_of_files, make_data_dicts, \
    range_str

from ..utils.mat import convert_mat_to_spect


def dataset(config_file):
    if not config_file.endswith('.ini'):
        raise ValueError('{} is not a valid config file, '
                         'must have .ini extension'.format(config_file))
    if not os.path.isfile(config_file):
        raise FileNotFoundError('config file {} is not found'
                                .format(config_file))
    config = ConfigParser()
    config.read(config_file)

    logger = logging.getLogger(__name__)
    logger.setLevel('INFO')

    labelset = config['DATA']['labelset']
    # make mapping from syllable labels to consecutive integers
    # start at 1, because 0 is assumed to be label for silent gaps
    if '-' in labelset or ',' in labelset:
        # if user specified range of ints using a str
        labelset = range_str(labelset)
    else:  # assume labelset is characters
        labelset = list(labelset)

    # to make type-checking consistent across .mat / .cbin / Koumura .wav files
    # set all_labels_are_int flag
    # currently only used with .mat files
    if config.has_option('DATA','all_labels_are_int'):
        all_labels_are_int = config.getboolean('DATA','all_labels_are_int')
    else:
        all_labels_are_int = False

    # map labels to series of consecutive integers from 0 to n inclusive
    # where 0 is the label for silent periods between syllables
    # and n is the number of syllable labels
    if all_labels_are_int:
        labels_mapping = dict(zip([int(label) for label in labelset],
                                  range(1, len(labelset) + 1)))
    else:
        labels_mapping = dict(zip(labelset,
                                  range(1, len(labelset) + 1)))
    if not config.has_option('DATA', 'silent_gap_label'):
        labels_mapping['silent_gap_label'] = 0
    else:
        labels_mapping['silent_gap_label'] = int(
            config['DATA']['silent_gap_label'])
    if sorted(labels_mapping.values()) != list(range(len(labels_mapping))):
        raise ValueError('Labels mapping does not map to a consecutive'
                         'series of integers from 0 to n (where 0 is the '
                         'silent gap label and n is the number of syllable'
                         'labels).')

    skip_files_with_labels_not_in_labelset = config.getboolean(
        'DATA',
        'skip_files_with_labels_not_in_labelset')

    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    if config.has_option('DATA', 'output_dir'):
        output_dir = os.path.join(config['DATA']['output_dir'],
                                  'spectrograms_' + timenow)
    else:
        output_dir = None

    ### if using spectrograms from .mat files ###
    if config.has_option('DATA','mat_spect_files_path'):
        # make spect_files file from .mat spect files and annotation file
        mat_spect_files_path = config['DATA']['mat_spect_files_path']
        print('will use spectrograms from .mat files in {}'
              .format(mat_spect_files_path))
        mat_spect_files = glob(os.path.join(mat_spect_files_path,'*.mat'))
        mat_spects_annotation_file = config['DATA']['mat_spect_files_annotation_file']
        if output_dir is None:
            output_dir = os.path.join(mat_spect_files_path,
                                  'spectrograms_' + timenow)
        os.mkdir(output_dir)

        spect_files_path = convert_mat_to_spect(mat_spect_files,
                                                mat_spects_annotation_file,
                                                output_dir,
                                                labels_mapping=labels_mapping)
    else:
        mat_spect_files_path = None

    ### if **not** using spectrograms from .mat files ###
    if mat_spect_files_path is None:
        if not config.has_section('SPECTROGRAM'):
            raise ValueError('No annotation_path specified in config_file that '
                             'would point to annotated spectrograms, but no '
                             'parameters provided to generate spectrograms '
                             'either.')
        # require user to specify parameters for spectrogram
        # instead of having defaults (as was here previously)
        # helps ensure we don't mix up different params
        spect_params = {}
        spect_params['fft_size'] = int(config['SPECTROGRAM']['fft_size'])
        spect_params['step_size'] = int(config['SPECTROGRAM']['step_size'])
        spect_params['freq_cutoffs'] = [float(element)
                                        for element in
                                        config['SPECTROGRAM']['freq_cutoffs']
                                            .split(',')]
        if config.has_option('SPECTROGRAM', 'thresh'):
            spect_params['thresh'] = float(config['SPECTROGRAM']['thresh'])
        if config.has_option('SPECTROGRAM', 'transform_type'):
            spect_params['transform_type'] = config['SPECTROGRAM']['transform_type']
            valid_transform_types = {'log_spect', 'log_spect_plus_one'}
            if spect_params['transform_type'] not in valid_transform_types:
                raise ValueError('Value for `transform_type`, {}, in [SPECTROGRAM] '
                                 'section of .ini file is not recognized. Must be one '
                                 'of the following: {}'
                                 .format(spect_params['transform_type'],
                                         valid_transform_types))

        data_dir = config['DATA']['data_dir']
        if not os.path.isdir(data_dir):
            raise NotADirectoryError('{} not recognized '
                                     'as a directory'.format(data_dir))
        logger.info('will make training data from: {}'.format(data_dir))
        if output_dir is None:
            output_dir = os.path.join(data_dir,
                                  'spectrograms_' + timenow)
        os.mkdir(output_dir)

        cbins = glob(os.path.join(data_dir, '*.cbin'))
        if cbins == []:
            # if we don't find .cbins in data_dir, look in sub-directories
            cbins = []
            subdirs = glob(os.path.join(data_dir, '*/'))
            for subdir in subdirs:
                cbins.extend(glob(os.path.join(data_dir,
                                               subdir,
                                               '*.cbin')))
        if cbins == []:
            # try looking for .wav files
            wavs = glob(os.path.join(data_dir, '*.wav'))

            if cbins == [] and wavs == []:
                raise FileNotFoundError('No .cbin or .wav files found in {} or'
                                        'immediate sub-directories'
                                        .format(data_dir))
            # look for canary annotation
            annotation_file = glob(os.path.join(data_dir, '*annotation*.mat'))
            if len(annotation_file) == 1:
                annotation_file = annotation_file[0]
            else:  # try Koumura song annotation
                annotation_file = glob(os.path.join(data_dir, '../Annotation.xml'))
                if len(annotation_file) == 1:
                    annotation_file = annotation_file[0]
                else:
                    raise ValueError('Found more than one annotation.mat file: {}. '
                                     'Please include only one such file in the directory.'
                                     .format(annotation_file))

        if cbins:
            spect_files_path = \
                make_spects_from_list_of_files(cbins,
                                               spect_params,
                                               output_dir,
                                               labels_mapping,
                                               skip_files_with_labels_not_in_labelset)
        elif wavs:
            spect_files_path = \
                make_spects_from_list_of_files(wavs,
                                               spect_params,
                                               output_dir,
                                               labels_mapping,
                                               skip_files_with_labels_not_in_labelset,
                                               annotation_file)

    saved_data_dict_paths = make_data_dicts(output_dir,
                float(config['DATA']['total_train_set_duration']),
                float(config['DATA']['validation_set_duration']),
                float(config['DATA']['test_set_duration']),
                labelset,
                spect_files_path)

    # lastly rewrite config file,
    # so that paths where results were saved are automatically in config
    for key, saved_data_dict_path in saved_data_dict_paths.items():
        config.set(section='TRAIN',
                   option=key + '_data_path',
                   value=saved_data_dict_path)
    with open(config_file, 'w') as config_file_rewrite:
        config.write(config_file_rewrite)

if __name__ == "__main__":
    config_file = os.path.normpath(sys.argv[1])
    dataset(config_file)



import os
from glob import glob
from datetime import datetime
import pickle

import joblib
import numpy as np
import tensorflow as tf
from tqdm import tqdm

from ..utils.data import reshape_data_for_batching, make_spects_from_list_of_files
from ..utils.mat import convert_mat_to_spect
from .. import network


def predict(checkpoint_path,
            networks,
            labels_mapping_path,
            spect_params,
            dir_to_predict=None,
            mat_spect_files_path=None,
            spect_scaler_path=None
            ):
    """make predictions with one trained model

    Parameters
    ----------
    checkpoint_path : str
        path to directory with saved model
    networks : dict
        where each key is the name of a neural network and the corresponding
        value is the configuration for that network (in a namedtuple or a dict)
    labels_mapping_path : str
        path to file that contains labels mapping, to convert output from consecutive
        digits back to labels used for audio segments (e.g. birdsong syllables)
    spect_params : dict
        Dictionary of parameters for creating spectrograms.
    dir_to_predict : str
        path to directory where input files are located
    mat_spect_files_path
        path to directory with .mat files containing spectrograms that should be used
        as data for which predictions are made
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
        Default is None.

    Returns
    -------
    None
    """
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')

    if dir_to_predict is None and mat_spect_files_path is None:
        raise ValueError('must specify either dir_to_predict or mat_spect_files_path but both are None')

    if dir_to_predict and mat_spect_files_path:
        raise ValueError('got values for both dir_to_predict and mat_spect_files_path, not clear which to use')

    with open(labels_mapping_path, 'rb') as labels_map_file_obj:
        labels_mapping = pickle.load(labels_map_file_obj)
    n_syllables = len(labels_mapping)

    if mat_spect_files_path:
        print('will use spectrograms from .mat files in {}'
              .format(mat_spect_files_path))
        mat_spect_files = glob(os.path.join(mat_spect_files_path, '*.mat'))
        spects_dir = os.path.join(mat_spect_files_path,
                                  'spectrograms_' + timenow)
        os.mkdir(spects_dir)
        spect_files_path = convert_mat_to_spect(mat_spect_files,
                                                mat_spects_annotation_file,
                                                spects_dir,
                                                labels_mapping=labels_mapping)
        dir_to_predict = spect_files_path

    else:
        if not os.path.isdir(dir_to_predict):
            raise FileNotFoundError('directory {}, specified as '
                                    'dir_to_predict, is not found.'
                                    .format(dir_to_predict))

        spects_dir = os.path.join(dir_to_predict,
                              'spectrograms_' + timenow)
        os.mkdir(spects_dir)

        cbins = glob(os.path.join(dir_to_predict, '*.cbin'))
        if cbins == []:
            # if we don't find .cbins in data_dir, look in sub-directories
            cbins = []
            subdirs = glob(os.path.join(dir_to_predict, '*/'))
            for subdir in subdirs:
                cbins.extend(glob(os.path.join(dir_to_predict,
                                               subdir,
                                               '*.cbin')))
        if cbins == []:
            # try looking for .wav files
            wavs = glob(os.path.join(dir_to_predict, '*.wav'))

            if cbins == [] and wavs == []:
                raise FileNotFoundError('No .cbin or .wav files found in {} or'
                                        'immediate sub-directories'
                                        .format(dir_to_predict))

        if cbins:
            spect_files_path = \
                make_spects_from_list_of_files(cbins,
                                               spect_params,
                                               spects_dir,
                                               labels_mapping,
                                               skip_files_with_labels_not_in_labelset=False,
                                               is_for_predict=True)
        elif wavs:
            spect_files_path = \
                make_spects_from_list_of_files(wavs,
                                               spect_params,
                                               spects_dir,
                                               labels_mapping,
                                               skip_files_with_labels_not_in_labelset=False,
                                               is_for_predict=True)

    # TODO should be able to just call dataset here, right?
    # instead of forcing user to specify spect_file_list
    # should give them option to do either
    spect_file_list = glob(os.path.join(spects_dir,
                                        '*.spect'))
    if spect_file_list == []:
        raise ValueError('did not find any .spect files in {}'
                         .format(dir_to_predict))

    X_data = []
    for spect_file in spect_file_list:
        spect_dict = joblib.load(spect_file)
        X_data.append(spect_dict['spect'].T)
    freq_bins_all_spects = [Xd.shape[-1] for Xd in X_data]
    uniq_freq_bins = set(freq_bins_all_spects)
    if len(uniq_freq_bins) != 1:
        raise ValueError('Found spectrograms with different numbers of frequency bins in dir_to_predict.\n'
                         f'Different values for numbers of frequency bins found were: {uniq_freq_bins}')
    else:
        freq_bins = list(uniq_freq_bins)[0]

    if not os.path.isdir(checkpoint_path):
        raise FileNotFoundError('directory {}, specified as '
                                'checkpoint_path, is not found.'
                                .format(checkpoint_path))
    checkpoint_file = tf.train.latest_checkpoint(checkpoint_path)
    meta_file = glob(checkpoint_file + '*meta*')
    if len(meta_file) > 1:
        raise ValueError('found more than one .meta file in {}'
                         .format(checkpoint_path))
    elif len(meta_file) < 1:
        raise ValueError('did not find .meta file in {}'
                         .format(checkpoint_path))
    else:
        meta_file = meta_file[0]

    data_file = glob(checkpoint_file + '*data*')
    if len(data_file) > 1:
        raise ValueError('found more than one .data file in {}'
                         .format(checkpoint_dir))
    elif len(data_file) < 1:
        raise ValueError('did not find .data file in {}'
                         .format(checkpoint_dir))
    else:
        data_file = data_file[0]

    num_spect_files = len(spect_file_list)

    NETWORKS = network._load()

    for net_name, net_config in networks.items():
        net_config_dict = net_config._asdict()
        net_config_dict['n_syllables'] = n_syllables
        if 'freq_bins' in net_config_dict:
            net_config_dict['freq_bins'] = freq_bins
        net = NETWORKS[net_name](**net_config_dict)

        if spect_scaler_path:
            spect_scaler = joblib.load(spect_scaler_path)

        with tf.Session(graph=net.graph) as sess:
            tf.logging.set_verbosity(tf.logging.ERROR)

            net.restore(sess=sess,
                        meta_file=meta_file,
                        data_file=data_file)

            preds_dict = {}
            pbar = tqdm(spect_file_list)
            for file_num, spect_file in enumerate(pbar):
                pbar.set_description(
                    f'Predicting labels for {os.path.basename(spect_file)}, file {file_num} of {num_spect_files}'
                )

                Xd = X_data[file_num]
                if spect_scaler_path:
                    Xd = spect_scaler.transform(Xd)
                (Xd_batch,
                 num_batches) = reshape_data_for_batching(Xd,
                                                          batch_size=net_config.batch_size,
                                                          time_steps=net_config.time_bins)

                if 'Y_pred' in locals():
                    del Y_pred
                # work through current spectrogram batch by batch
                for b in range(num_batches):  # "b" is "batch number"
                    d = {net.X: Xd_batch[:, b * net_config.time_bins: (b + 1) * net_config.time_bins, :],
                         net.lng: [net_config.time_bins] * net_config.batch_size}
                    if 'Y_pred' in locals():
                        # if Y_pred exists, we concatenate with new predictions
                        # for next batch
                        preds = sess.run(net.predict, feed_dict=d)
                        preds = preds.reshape(-1)  # batch_size
                        Y_pred = np.concatenate((Y_pred, preds), axis=0)
                    else:  # if Y_pred doesn't exist yet
                        Y_pred = sess.run(net.predict, feed_dict=d)
                        Y_pred = Y_pred.reshape(-1)

                # remove zero padding added by reshape_data_for_batching function
                Y_pred = Y_pred.ravel()
                Y_pred = Y_pred[0:Xd.shape[0]]
                preds_dict[spect_file] = Y_pred

        fname = os.path.join(dir_to_predict, 'predictions')
        joblib.dump(preds_dict, fname)



import os
from configparser import ConfigParser
from configparser import MissingSectionHeaderError, ParsingError, DuplicateOptionError, DuplicateSectionError

from .train import train
from .predict import predict
from .learncurve import learncurve
from .summary import summary
from .prep import make_data
from ..config import parse_spect_config, parse_data_config, parse_train_config, \
    parse_predict_config, parse_output_config
from ..config.parse import _get_nets_config


def cli(command, config_files):
    """command-line interface

    Parameters
    ----------
    command : string
        One of {'prep', 'train', 'predict', 'finetune', 'learncurve', 'summary'}
    config_files : string or list
        config.ini files
    """
    for config_file in config_files:
        # check config_file exists,
        # because if it doesn't ConfigParser will just return an "empty" instance w/out sections or options
        if not os.path.isfile(config_file):
            raise FileNotFoundError(f'config file not found: {config_file}')

        try:
            config_obj = ConfigParser()
            config_obj.read(config_file)
        except (MissingSectionHeaderError, ParsingError, DuplicateOptionError, DuplicateSectionError):
            # try to add some context for users that do not spend their lives thinking about ConfigParser objects
            print(f"Error when opening the following config_file: {config_file}")
            raise
        except:
            # say something different if we can't add very good context
            print(f"Unexpected error when opening the following config_file: {config_file}")
            raise

        if command == 'prep':
            data_config = parse_data_config(config_obj, config_file)
            spect_params = parse_spect_config(config_obj)
            if spect_params is None and data_config.mat_spect_files_path is None:
                # then user needs to specify spectrogram parameters
                raise ValueError('No annotation_path specified in config_file that '
                                 'would point to annotated spectrograms, but no '
                                 'parameters provided to generate spectrograms '
                                 'either.')

            make_data(labelset=data_config.labelset,
                      all_labels_are_int=data_config.all_labels_are_int,
                      data_dir=data_config.data_dir,
                      total_train_set_dur=data_config.total_train_set_dur,
                      val_dur=data_config.val_dur,
                      test_dur=data_config.test_dur,
                      config_file=config_file,
                      silent_gap_label=data_config.silent_gap_label,
                      skip_files_with_labels_not_in_labelset=data_config.skip_files_with_labels_not_in_labelset,
                      output_dir=data_config.output_dir,
                      mat_spect_files_path=data_config.mat_spect_files_path,
                      mat_spects_annotation_file=data_config.mat_spects_annotation_file,
                      spect_params=spect_params)

        elif command == 'train':
            train_config = parse_train_config(config_obj, config_file)
            nets_config = _get_nets_config(config_obj, train_config.networks)
            spect_params = parse_spect_config(config_obj)
            output_config = parse_output_config(config_obj)
            train(train_data_dict_path=train_config.train_data_dict_path,
                  val_data_dict_path=train_config.val_data_dict_path,
                  spect_params=spect_params,
                  networks=nets_config,
                  num_epochs=train_config.num_epochs,
                  config_file=config_file,
                  val_error_step=train_config.val_error_step,
                  checkpoint_step=train_config.checkpoint_step,
                  patience=train_config.patience,
                  save_only_single_checkpoint_file=train_config.save_only_single_checkpoint_file,
                  normalize_spectrograms=train_config.normalize_spectrograms,
                  root_results_dir=output_config.root_results_dir,
                  save_transformed_data=train_config.save_transformed_data)

        elif command == 'finetune':
            raise NotImplementedError

        elif command == 'predict':
            predict_config = parse_predict_config(config_obj)
            nets_config = _get_nets_config(config_obj, predict_config.networks)
            spect_params = parse_spect_config(config_obj)
            predict(checkpoint_path=predict_config.checkpoint_path,
                    networks=nets_config,
                    labels_mapping_path=predict_config.labels_mapping_path,
                    spect_params=spect_params,
                    dir_to_predict=predict_config.dir_to_predict,
                    spect_scaler_path=predict_config.spect_scaler_path)

        elif command == 'learncurve':
            train_config = parse_train_config(config_obj, config_file)
            nets_config = _get_nets_config(config_obj, train_config.networks)
            spect_params = parse_spect_config(config_obj)
            data_config = parse_data_config(config_obj, config_file)
            output_config = parse_output_config(config_obj)
            if train_config.train_data_dict_path is None:
                raise ValueError("must set 'train_data_path' option in [TRAIN] section of config.ini file "
                                 "before running 'learncurve'")
            if train_config.val_data_dict_path is None:
                raise ValueError("must set 'val_data_path' option in [TRAIN] section of config.ini file "
                                 "before running 'learncurve'")
            learncurve(train_data_dict_path=train_config.train_data_dict_path,
                       val_data_dict_path=train_config.val_data_dict_path,
                       spect_params=spect_params,
                       total_train_set_duration=data_config.total_train_set_dur,
                       train_set_durs=train_config.train_set_durs,
                       num_replicates=train_config.num_replicates,
                       num_epochs=train_config.num_epochs,
                       config_file=config_file,
                       networks=nets_config,
                       val_error_step=train_config.val_error_step,
                       checkpoint_step=train_config.checkpoint_step,
                       patience=train_config.patience,
                       save_only_single_checkpoint_file=train_config.save_only_single_checkpoint_file,
                       normalize_spectrograms=train_config.normalize_spectrograms,
                       use_train_subsets_from_previous_run=train_config.use_train_subsets_from_previous_run,
                       previous_run_path=train_config.previous_run_path,
                       root_results_dir=output_config.root_results_dir,
                       save_transformed_data=data_config.save_transformed_data)

        elif command == 'summary':
            train_config = parse_train_config(config_obj, config_file)
            nets_config = _get_nets_config(config_obj, train_config.networks)
            data_config = parse_data_config(config_obj, config_file)
            output_config = parse_output_config(config_obj)
            summary(results_dirname=output_config.results_dirname,
                    train_data_dict_path=train_config.train_data_dict_path,
                    networks=nets_config,
                    train_set_durs=train_config.train_set_durs,
                    num_replicates=train_config.num_replicates,
                    labelset=data_config.labelset,
                    test_data_dict_path=train_config.test_data_dict_path,
                    normalize_spectrograms=train_config.normalize_spectrograms)



"""validators used by attrs-based classes"""
import os


def is_a_directory(instance, attribute, value):
    """check if given path is a directory"""
    if not os.path.isdir(value):
        raise NotADirectoryError(
            f'Value specified for {attribute.name} of {type(instance)} not recognized as a directory:\n'
            f'{value}'
        )


def is_a_file(instance, attribute, value):
    """check if given path is a directory"""
    if not os.path.isfile(value):
        raise NotADirectoryError(
            f'Value specified for {attribute.name} of {type(instance)} not recognized as a file:\n'
            f'{value}'
        )
